import sys
import argparse
import gzip
import io
from pgpipe.vcf_reader_func import VcfReader
from pgpipe.model import Model, read_single_model

def createParser():
    parser = argparse.ArgumentParser(description=("Outputs windows in a "
                      "VCF file where all sites in the window have below"
                      " a set limit of missing data. Limit default "
                      "is any missing data."))
    parser.add_argument('--vcf',dest="vcfname",type=str,
                        help=("Name of input VCF file, default is stdin"))
    parser.add_argument('--size',dest="window_size",type=int,default=1000)
    parser.add_argument('--zero-ho',dest="zero_ho",action="store_true",
                        help=("Output list in zero-based half-open format"))
    parser.add_argument('--zero-closed',dest="zero_closed",action="store_true",
                        help=("Output list in zero-based closed format"))
    parser.add_argument('--extend-regions',dest="extend",action="store_true",
                        help=("Regions will be cut at halfway point between "
                        "valid and invalid site, instead of at site"))
    parser.add_argument('--missing-count',dest="missing_count",type=int,
                        help=("Count variants with more than (n) missing "
                        "samples as a missing site"),default=0)
    parser.add_argument('--out',dest='outname',type=str,
                        help="Output filename (default is stdout)")
    parser.add_argument('--addchr',dest="addchr",action="store_true",
                        help=("Add 'chr' to start of chromosome name"))
    parser.add_argument('--removechr',dest="removechr",action="store_true",
                        help=("Remove 'chr' from start of chromosome name"))
    parser.add_argument('--model-file',dest="popname",help="Model file if only "
                        "using subset of inds in a file")
    parser.add_argument('--model',dest="modelname",help=("Name of population "
                        "in model file if more than one is contained"))
    parser.add_argument("--tbi", dest="tabix_index", help="Path to bgzipped "
                        "file's index if name doesn't match VCF file")
    return parser

def getl(f,compressed=False):
    l = f.readline()
    if compressed:
        return l.decode()
    return l

def fixChromName(chrom,addchr,removechr):
    if addchr and chrom[3:] != 'chr':
        return 'chr'+chrom
    if removechr and chrom[3:] == 'chr':
        return chrom[3:]
    return chrom

def outputLine(chrom,start_pos,end_pos,args):
    sp = start_pos
    ep = end_pos
    if args.zero_ho:
        sp -= 1
    elif args.zero_closed:
        sp -= 1
        ep -= 1
    chrom = fixChromName(chrom,args.addchr,args.removechr)
    return chrom+'\t'+str(sp)+'\t'+str(ep)+'\n'

def regionsWithData(sysargs):
    """Returns a BED file with regions where all SNPs contained have
    less than a given limit of missing data.

    Given an input VCF file (regular,bgzipped,stdin), will output regions
    where all SNPs in the region have under a certain threshhold of missing
    data, as set by --missing-count. The value specified here will cut
    off regions where more than 'n' individuals are missing data, the 
    default value for this is 0. Options are included to subsample VCF by
    individual and only track missing data for those included, adding or 
    dropping 'chr' form the chromosome name, and setting a minimum size for
    the length of output windows. 

    Parameters
    ----------
    --vcf : str
        Filename for VCF input file. If not provided, defaults to stdin
    --size : int (1000)
        Outputs regions that have length equal to or greater than 
        provided value
    --missing-count : int (0)
        Any site with more than (n) individuals missing data will be 
        considered as a region breakpoint. If used with --model-file,
        will only count individuals that are included in model.
    --model-file : str
        Used to subsample VCF and only look at missing data in given
        individuals
    --model : str
        If model file has more than one model, the model name is passed
        here to specify which one to use.
    --out : str
        Output BED filename
    
    Other Parameters
    ----------------
    --zero-ho : bool
        If set, treats regions in BED file as 0-based, (h)alf (o)pen
        coordinates rather than 1-based, closed.
    --extend-regions : bool
        If set, will extend regions past the flanking included SNPs
        to the halfway point between the end SNP and the SNP with
        missing data
    --addchr : bool
        If set, will add 'chr' to any chromosome name
    --removechr : bool
        If set, will remove 'chr' from any chromosome name that starts with it
    --tbi : str
        Name of tabix index for VCF if using bgzipped file without default
        tbi index name
    


    """
    parser = createParser()
    args = parser.parse_args(sysargs)
    compressed_input = False
    #if args.vcfname is None:
    #    instream = sys.stdin
    #elif args.vcfname[-3:] == '.gz':
    #    compressed_input = True
    #    instream = gzip.open(args.vcfname,'r')
   # else:
   #     instream = open(args.vcfname,'r')


    if args.vcfname is None:
        vcfname = '-'
    else:
        vcfname = args.vcfname

    popmodel = None
    if args.popname is not None:
        popmodel = read_single_model(args.popname,args.modelname)

    vcf_in = VcfReader(vcfname,
                       popmodel=popmodel,
                       index=args.tabix_index)

    if args.outname is not None:
        outstream = open(args.outname,'w')
    else:
        outstream = sys.stdout

    #sample_count = 20
    #missing_data_count = [0 for i in range(sample_count+1)]
    sample_count = 0
    indel_count = 0

    #missing_data_per_indiv = [0 for i in range(sample_count)]

    prev_miss_pos = 0
    prev_full_pos = 0
    prev_pos = 0
    prev_chrom = ''

    in_section = False
    #May be skipping first site, fine here but issue elsewhere
    for record in vcf_in.reader:
        missing_for_site = False
        missing_at_site = 0
        #for i in range(9,len(la)):
        for i in range(len(record.samples)):
            #geno = la[i]
            #if '.' in geno:
            if record.samples[i].alleles[0] in [None,'N']:
                missing_at_site += 1
                if missing_at_site > args.missing_count:
                    missing_for_site = True
                    break
        #cur_pos = int(la[1])
        cur_pos = record.pos
        #if prev_pos == 0 or prev_chrom != la[0]:
        if prev_pos == 0 or prev_chrom != record.chrom:
            if prev_pos != 0:
                start_pos = ((prev_miss_pos+prev_full_pos+1)//2 if args.extend else prev_full_pos)
                end_pos = prev_pos
                outstream.write(outputLine(prev_chrom,start_pos,end_pos,args))
            prev_miss_pos = cur_pos
            prev_full_pos = cur_pos
            #prev_chrom = la[0]
            prev_chrom = record.chrom
            if not missing_for_site:
                in_section = True
            prev_pos = cur_pos
            continue
        if missing_for_site:
            if in_section:
                in_section = False
                if args.extend:
                    start_pos = (prev_miss_pos+prev_full_pos+1)//2
                    end_pos = (prev_pos+cur_pos)//2
                else:
                    start_pos = prev_full_pos
                    end_pos = prev_pos
                diff = end_pos - start_pos
                if diff > args.window_size:
                    outstream.write(outputLine(record.chrom,start_pos,end_pos,args))
        else:
            if not in_section:
                prev_miss_pos = prev_pos
                prev_full_pos = cur_pos
                in_section = True
        prev_pos = cur_pos
        #line = getl(instream,compressed_input)
    if in_section:
        if args.extend:
            start_pos = (prev_miss_pos+prev_full_pos+1)//2
        else:
            start_pos = prev_full_pos
        end_pos = prev_pos
        outstream.write(outputLine(prev_chrom,start_pos,end_pos,args))
    try:
        outstream.close()
    except:
        pass

if __name__ == '__main__':
    regionsWithData(sys.argv[1:])



import sys
import re
import logging
from functools import total_ordering
from random import shuffle


def getChromKey(chrom):
    """Get key from chromosome string so that it can be naturally sorted
    (int < string, numbers grouped together)
    """
    c = chrom
    if chrom[0:3] == 'chr':
        c = chrom[3:]
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    k = [ convert(i) for i in re.split('([0-9]+)',c) if len(i) != 0]
    return k

def keyComp(k1,k2):
    """Returns True if k1 < k2, False if not. int < str"""
    for i in range(min(len(k1),len(k2))):
        e1 = k1[i]
        e2 = k2[i]
        if isinstance(e1,int) != isinstance(e2,int):
            return isinstance(e1,int)
        if e1 != e2:
            return e1 < e2
    return len(k1) < len(k2)

def setRegionSort(sorttype,sortlist=None):
    valid_types = ['natural','string','list']
    if sorttype not in valid_types:
        raise Exception("Sorting type %s is not valid" % (sorttype))
    Region.sort_method = sorttype
    if sorttype == 'list' and sortlist is not None:
        setSortOrder(sortlist)

def setSortOrder(sortlist,trimchrom=False):
    hold_list = []
    for c in sortlist:
        if trimchrom and c[0:3] == 'chr':
            hold_list.append(c[3:])
        else:
            hold_list.append(c)
    Region.sort_order = hold_list

@total_ordering
class Region:
    sort_method = "natural"
    sort_order = None

    def __init__(self, start, end, chrom, fullline = None):
        """Zero-based, half open coordinates and chromosome info for
        a region in a genome. Coords will be formatted according to
        flags passed to RegionList, then stored in a single format.
        """
        self.start = start
        self.end = end
        self.chrom = chrom
        self.fullline = fullline
        #if chrom[0:3] == 'chr':
        #    self.chrom = chrom[3:]
        #else:
        #    self.chrom = chrom
    def __deepcopy__(self):
        return Region(self.start,self.end,self.chrom)

    def cloneRegion(self):
        return Region(self.start,self.end,self.chrom)


    def getChromKey(self):
        """Splits chromosone name for natural sort. Ints and strs are
        grouped with adjacent elements of the same type

        Example: 'front88place' returns ['front',88,'place']
        """
        return getChromKey(self.chrom)


    def __eq__(self, other):
        return ((self.start == other.start) and (self.end == other.end)
                and (self.chrom == other.chrom))

    def __lt__(self, other):
        """Sort key order: chrom-key, start position, end position
        """
        if Region.sort_method == 'natural':
            k1 = self.getChromKey()
            k2 = other.getChromKey()
            if k1 != k2:
                return keyComp(k1,k2)
        elif Region.sort_method == "string":
            #k1 = self.chrom
            #k2 = other.chrom
            k1 = (self.chrom if self.chrom[:3] != 'chr' else self.chrom[3:])
            k2 = (other.chrom if other.chrom[:3] != 'chr' else other.chrom[3:])
            if k1 != k2:
                return k1 < k2
        else:
            if Region.sort_order is None:
                raise Exception("Sorting by list requires calling setSortOrder() and providing a list for the order")
            k1 = self.chrom
            k2 = other.chrom
            if k1 not in Region.sort_order:
                raise Exception("Key %s is not in sort order list" % (k1))
            if k2 not in Region.sort_order:
                raise Exception("Key %s is not in sort order list" % (k2))
            if k1 != k2:
                return Region.sort_order.index(k1) < Region.sort_order.index(k2)
        if self.start != other.start:
            return self.start < other.start
        return self.end < other.end

    def __len__(self):
        return self.end-self.start

    def containsRecord(self, rec):
        k1 = self.getChromKey()
        kr = getChromKey(rec.chrom)
        if k1 != kr:
            if keyComp(k1,kr):
                return 'before'
            return 'after'
        comp_pos = rec.pos-1
        if comp_pos < self.start:
            return 'before'
        if comp_pos >= self.end:
            return 'after'
        return 'in'

    def toStr(self, zeroho=False, zeroclosed=False, sep=':'):
        if self.fullline is not None:
            return self.fullline
        start = self.start
        end = self.end
        if not zeroho:
            start += 1
        elif zeroclosed:
            end -= 1
        return str(self.chrom)+sep+str(start)+sep+str(end)
        #return str(start)+sep+str(end)+sep+str(self.chrom)

    def getReference(self, refseq):
        return refseq.fetch(self.chrom,self.start,self.end)




class RegionList:

    def __init__(self, filename=None, genestr=None, reglist=None,
                 zeroclosed=False, zeroho=False,
                 colstr=None, sortlist=True, checkoverlap=None,
                 sortmethod=None, sortorder=None, chromfilter=None,
                 list_template=None, randomize=False,keep_full_line=False):
        """Class for storing gene region information

        Will create a list that stores genomic regions imported from
        a file, string, or list. List can be sorted in various ways,
        filtered for regions on one or multiple chromosomes, convert
        between coordinate systems, and integrated with other PPP
        functions.

        Parameters
        ----------
        filename : str (None)
            Name of file with gene coordinate data
        genestr : str (None)
            Semicolon-separated list of region data, either in "start:end"
            or "start:end:chrom" format.
        zeroho : bool (False)
            If true, input is in zero-based, half-open coordinates. Since
            pysam uses this format, coordinates will be unaltered
            (default is to use one-based closed coordinates)
        zeroclosed : bool (False)
            If true, uses zero-based, closed coordinates, which adds 1
            to the end value for a region. Rarely if ever used option.
        colstr : str (None)
            Three-element string separated by semicolons, with
            each element being an integer corresponding to the 0-based index of the column of the start, end, and chromosome values
            for input region(s).
        sortlist : bool (True)
            Will sort list once initialized using assigned sorting scheme.
        checkoverlap : str (None, 'error','fix')
            If None, will not check for overlaps. If 'error', will raise
            an exception if overlap is found. If 'fix', overlapping
            regions will be merged. Note zero-based, half-open intervals
            do not overlap when adjacent ones share end/start position,
            but one-based closed intervals do.
        sortmethod : str (None,'natural','string','list')
            Determines method for sorting. 'natural' is pre-set method,
            passing None to the constructor will ensure sort type is
            unchanged from other calls. This sorts akin to 'sort -V' unix
            command, which breaks a string into ints and substrings for
            sorting. 'string' will sort via regular string method. 'list'
            will use list passed to sortorder as ordering of chromosomes.
        sortorder : list (None)
            If 'list' sortmethod is used, regions will be sorted according
            to the order specified in this list.
        chromfilter : str (None)
            If set, will allow only regions from passed chromosome
            into the region list.
        list_template : RegionList (None)
            If creating new list with regions from a different list,
            will use options from given list instead of arguments passed
            for the following arguments: zeroho, zeroclosed, sortlist,
            checkoverlap, and sortorder.
        randomize : bool (False)
            If set, will randomly shuffle regions in list.

        Exceptions
        ----------
        Both filename and genestr are None
        colstr has less than 2 or more than 3 values

        """
        self.collist = None
        if list_template is not None:
            zeroho = list_template.zeroho
            zeroclosed = list_template.zeroclosed
            sortlist = list_template.sortlist
            checkoverlap = list_template.checkoverlap
            sortorder = list_template.sortorder

        if self.collist is None:
            if colstr is None:
                self.collist = [1, 2, 0]
            else:
                self.parseCols(colstr)
        if filename is None and genestr is None and reglist is None:
            raise Exception(("A filename, region string, or list of "
                             "region strings must be provided for " "creating a gene region list"))
        if [filename,genestr,reglist].count(None) < 2:
            raise Exception(("Only one of filename, region string, and "
                            "region list can be specified for init"))
        if zeroho and zeroclosed:
            raise Exception(("Zero-halfopen and zero-closed options "
                             "cannot be invoked at the same time"))

        if sortlist and randomize:
            raise Exception("Sorting and randomizing are not compatible options for RegionList")
        if checkoverlap not in [None,'fix','error']:
            raise Exception("Invalid checkoverlap value: %s" % (checkoverlap))
        self.regions = []
        self.zeroho = zeroho
        self.zeroclosed = zeroclosed
        self.chromfilter = chromfilter
        self.keep_full_line = keep_full_line
        self.header = None
        if sortmethod is not None:
            setRegionSort(sortmethod,sortlist=sortorder)
        if filename is not None:
            self.initFile(filename)
        elif genestr is not None:
            self.initStr(genestr)
        else:
            self.initList(reglist)

        if sortlist:
            self.regions.sort()
        if checkoverlap is not None:
            if self.hasOverlap():
                if checkoverlap == "fix":
                    self.fixOverlap()
                elif checkoverlap == "error":
                    raise Exception("Region overlap detected")

        if randomize:
            shuffle(self.regions)

    def initFile(self,filename):
        """Initialize RegionList with a region file
        """
        past_header = False
        with open(filename, 'r') as regionfile:
            for line in regionfile:
                if line[0] == '#':
                    self.header = line.strip()
                    past_header = True
                    continue
                la = line.strip().split()
                if not past_header:
                    past_header = True
                    try:
                        t = int(la[self.collist[0]])
                    except ValueError as e:
                        self.header = line.strip()
                        continue
                self.initRegion(la)


    def initStr(self,genestr):
        la = genestr.split(':')
        self.initRegion(la)


    def initList(self,reglist):
        for reg in reglist:
            self.initRegion(reg)



    def initRegion(self,la):
        start = int(la[self.collist[0]])
        end = int(la[self.collist[1]])
        chrom = la[self.collist[2]]
        if self.chromfilter is not None and chrom != self.chromfilter:
            return
        if not self.zeroho:
            start -= 1
        elif self.zeroclosed:
            end += 1
        fullline = (('\t'.join(la)) if self.keep_full_line else None)
        self.regions.append(Region(start,end,chrom,fullline))


    def parseCols(self,cols):
        col_list = [int(i) for i in cols.split(',')]
        if len(col_list) != 3:
            raise Exception(("Column string %s must have three "
                             "values" % (cols)))
        self.collist = col_list

    def toStr(self, idx, sep=':'):
        return self.regions[idx].toStr(zeroho=self.zeroho,
                                       zeroclosed=self.zeroclosed,
                                       sep=sep)

    def hasOverlap(self):
        region_hold = sorted(self.regions)
        for i in range(len(region_hold)-1):
            if region_hold[i].chrom == region_hold[i+1].chrom:
                if (region_hold[i].start <= region_hold[i+1].start
                    < region_hold[i].end):
                    logging.warning("Regions %s and %s overlap" %
                                    (self.toStr(i),self.toStr(i+1)))
                    return True
        return False

    def fixOverlap(self):
        region_hold = []
        self.regions.sort()
        i = 0
        temp_reg = None
        for i in range(len(self.regions)):
            if temp_reg is None:
                temp_reg = self.regions[i].cloneRegion()
                continue
            cur_reg = self.regions[i].cloneRegion()
            if cur_reg.start < temp_reg.end and cur_reg.chrom == temp_reg.chrom:
                temp_reg.end = max(temp_reg.end,cur_reg.end)
            else:
                region_hold.append(temp_reg)
                temp_reg = cur_reg
        region_hold.append(temp_reg)
        self.regions = region_hold

    def printList(self, file_handle=None, file_name=None,
                  return_str=False, delim="\t", add_chr=False,
                  remove_chr=False):
        if file_handle is None and file_name is None:
            file_handle = sys.stdout
        if file_name is not None:
            file_handle = open(file_name, 'w')
        if return_str:
            out_str = ''
        if self.header is not None:
            file_handle.write(self.header+'\n')
        for region in self.regions:
            if region.fullline is not None:
                file_handle.write(region.fullline+'\n')
                continue
            start = region.start
            end = region.end
            if not self.zeroho:
                start += 1
            elif self.zeroclosed:
                end -= 1
            chrom = region.chrom
            if add_chr and chrom[:3] != 'chr':
                chrom = "chr"+region.chrom
            if remove_chr and chrom[:3] == 'chr':
                chrom = region.chrom[3:]
            reg_str = chrom+delim+str(start)+delim+str(end)+'\n'
            if return_str:
                out_str += reg_str
            else:
                file_handle.write(reg_str)
        if return_str:
            return out_str

    def filterByChrom(self,chrom_list,include=False):
        if include:
            self.regions = [r for r in self.regions if r.chrom in chrom_list]
        else:
            self.regions = [r for r in self.regions if r.chrom not in chrom_list]

    def filterOutXY(self):
        self.filterByChrom(['X','Y','chrX','chrY'])

    def expandRegions(self,expand_size):
        #Expands regions, but does not merge overlaps 
        #that may occur
        for i in range(len(self.regions)):
            self.regions[i].start = max(self.regions[i].start-expand_size,0)
            self.regions[i].end += expand_size

    #TO do:
    #Add sort method for strictly text based sorting

def getIntervalsBetween(region_list, padding=0, firstline=True):
    region_hold = []
    if len(region_list.regions) < 2:
        raise Exception("Region list for complement requires at least two regions")
    if firstline:
        r1 = region_list.regions[0]
        if r1.start - padding > 0:
            region_hold.append([str(r1.chrom),0,r1.start-padding])
    for i in range(len(region_list.regions)-1):
        r1 = region_list.regions[i]
        rn = region_list.regions[i+1]
        if r1.chrom != rn.chrom:
            continue
        new_start = r1.end+padding
        new_end = rn.start-padding
        if new_start < new_end:
            region_hold.append([str(r1.chrom),new_start,new_end])
    out_list = RegionList(reglist=region_hold, zeroho=True)
    out_list.zeroho = region_list.zeroho
    out_list.zeroclosed = region_list.zeroclosed
    return out_list

def subtractBed(stat_list, filter_list):
    #Sorts regions and removes regions from stat list 
    #that overlap with any region in filter_list
    stat_idx = 0
    filter_idx = 0
    stat_list.regions.sort()
    filter_list.regions.sort()
    drop_list = [False for i in stat_list.regions]
    stat_region = stat_list.regions[stat_idx]
    filter_region = filter_list.regions[filter_idx]
    while stat_idx < len(stat_list.regions):
        #try:
        #    print (stat_list.regions[stat_idx].toStr(),filter_list.regions[filter_idx].toStr())
        #except:
        #    break
        #print (stat_list.regions[stat_idx].toStr())
        while filter_idx < len(filter_list.regions) and stat_list.regions[stat_idx].end > filter_list.regions[filter_idx].start:
            #print (filter_list.regions[filter_idx].toStr())
            if stat_list.regions[stat_idx].start < filter_list.regions[filter_idx].end:
                #print ("drop "+str(stat_idx))
                drop_list[stat_idx] = True
                break
            filter_idx+=1
        stat_idx += 1
    stat_list.regions = [stat_list.regions[i] for i in range(len(stat_list.regions)) if drop_list[i] is False]
    return 



import sys
import pysam
import logging
import struct
from random import sample
from collections import defaultdict
import os
import gzip
import numpy as np


#sys.path.insert(0,os.path.abspath(os.path.join(os.pardir, 'andrew')))
from pgpipe.logging_module import initLogger
from pgpipe.model import read_model_file
from pgpipe import test_cython

def checkIfGzip(filename):
    try:
        gf = gzip.open(filename)
        gl = gf.readline()
        gf.close()
        vcf_check = b'##fileformat=VCF'
        if gl[0:3] == b'BCF':
            return 'bcf'
        elif gl[:len(vcf_check)] == vcf_check:
            return checkHeader(filename)
        else:
            return 'other'
    except OSError:
        return 'nozip'
    except IOError:
        return 'nozip'

def checkHeader(filename):
    f = open(filename,'rb')
    l = f.readline()
    f.close()
    BGZF_HEADER=b'\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00\x42\x43\x02\x00'
    #BGZF_HEADER=b'\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff'
    GZF_HEADER=b'\x1f\x8b'
    if l[:len(BGZF_HEADER)] == BGZF_HEADER:
        return 'bgzip'
    if l[:len(GZF_HEADER)] == GZF_HEADER:
        return 'gzip'
    return 'nozip'

def checkFormat(vcfname):
    """Checks header of given file for compression type


    Given a filename, opens file and reads first line to check if
    file has BGZF or GZIP header. May be extended to check for BCF format

    Parameters
    ----------
    filename : str
        Name of file to be checked

    Returns
    -------
    extension : str {'bgzip','gzip','vcf','other'}
        File extension as indicated by header

    """
    if vcfname == '-':
        return 'vcf' #may want to note its a stream
    typ = checkIfGzip(vcfname)
    if typ != 'nozip':
        return typ
    f = open(vcfname)
    l = f.readline()
    f.close()
    VCF_TAG='##fileformat=VCF'
    if l[:len(VCF_TAG)] == VCF_TAG:
        return 'vcf'
    return 'other'



def checkIfCpG(record,fasta_ref,add_chr=False):
    #Note: will not detect multiallelic/indel combos
    dr = None
    pos = record.pos
    c = record.chrom
    if record.alts is None:
        return False
    if add_chr:
        c = 'chr'+record.chrom
    if record.ref == 'C' and 'T' in record.alts:
        try:
            seq = fasta_ref.fetch(c,pos-1,pos+1)
        except KeyError:
            seq = fasta_ref.fetch(flipChrom(c),pos-1,pos+1)
        if seq[0].upper() != 'C':
            logging.warning('%s %d has bad base %s' % (record.chrom,record.pos,seq[0]))
        if seq[1].upper() == 'G':
            return True
        return False
    elif record.ref == 'G' and 'A' in record.alts:
        try:
            seq = fasta_ref.fetch(c,pos-2,pos)
        except KeyError:
            seq = fasta_ref.fetch(flipChrom(c),pos-2,pos)
        if seq[1].upper() != 'G':
            logging.warning('%s %d has bad base %s' % (record.chrom,record.pos,seq[1]))
        if seq[0].upper() == 'C':
            return True
        return False
    return False

def checkForDuplicates(rec_list,pass_list):
    for i in range(len(rec_list)-1):
        if rec_list[i].pos == rec_list[i+1].pos:
            pass_list[i] = False
            pass_list[i+1] = False

def checkForMultiallele(rec_list,pass_list):
    for i in range(len(rec_list)):
        if i != len(rec_list)-1 and rec_list[i].pos == rec_list[i+1].pos:
            pass_list[i] = False
            pass_list[i+1] = False
        if len(rec_list[i].alleles) > 2:
            pass_list[i] = False

def checkPopWithoutMissing(rec_list,model,pop_keys,min_per_pop=5):
    for pop in pop_keys:
        present_count = 0
        for i in range(len(pop_keys[pop])):
            tidx = pop_keys[pop][i]
            if tidx != -1:
                all_data = True
                for rec in rec_list:
                    rec_alleles = rec.samples[tidx].alleles
                    if rec_alleles[0] in [None,'N']:
                        all_data = False
                        break
                if all_data:
                    present_count += 1
                if present_count >= min_per_pop:
                    break
        if present_count < min_per_pop:
            return False
    return True

def readSinglePopModel(popfilename,popname=None):
    popmodels = read_model_file(popfilename)
    if len(popmodels) != 1:
        if popname is None:
            raise Exception("Model file %s needs one model to be specified")
        return popmodels[popname]
    else:
        pp = list(popmodels.keys())
        return popmodels[pp[0]]


def flipChrom(chrom):
    if chrom[0:3] == 'chr':
        return chrom[0:3]
    return 'chr'+chrom

def getAlleleCountCython(rec,idxlist=None):
    alleles = defaultdict(int)
    srec = str(rec)
    if idxlist is None:
        d_array = test_cython.getAlleleCountArray(srec.encode(),len(srec))
    else:
        d_array = test_cython.getAlleleCountArraySub(srec.encode(),len(srec),idxlist,len(idxlist))
    for i,a in enumerate(rec.alleles):
        alleles[a] = d_array[i+48]
    if d_array[46] != 0:
        alleles['N'] = d_array[46]
    return alleles

def getAlleleCountDict(rec,idx_list=None):
    """
    Returns dict with allele counts for all
    """
    alleles = defaultdict(int)
    total_sites = 0
    missing_inds = 0
    if idx_list is None:
        idx_list = range(len(rec.samples))
    for j in idx_list:
        samp = rec.samples[j]
        if None in samp.alleles:
            alleles['N'] += len(samp.alleles)
            #missing_inds += 1
        for k in range(len(samp.alleles)):
            b = samp.alleles[k]
            if b is not None:
                alleles[b] += 1
            total_sites+=1
    return alleles

def getAlleleStats(rec):
    acd = getAlleleCountDict(rec)
    missing_inds = (acd['N'] if 'N' in acd.keys() else 0)
    total_sites = sum(acd.values())-missing_inds
    return acd,total_sites,missing_inds

#def parseRecString(recstr):



def getAlleleStatsAlt(rec):
    ala = str(rec).strip().split()[9:]
    l = []
    allele_idx = rec.alleles
    d = {}
    for a in allele_idx:
        d[a] = 0
    d['.'] = 0
    for j in range(len(ala)):
        for k in ala[j].split(':')[0].split(ala[j][1]):
            #d[allele_idx]
            if k == '.':
                d['N'] += 1
            else:
                d[allele_idx[int(k)]] += 1
    #print (str(d))
    return d
        #d[allele_idx[ii]] += 1 for ii in ala[j].split(':').split('|'))
        #l.extend(ala.split(':')[0].split('|'))

    #npa = np.array(l)

def getAlleleStatsCython(rec,idx_list=None):
    acd = getAlleleCountCython(rec,idx_list)
    missing_inds = (acd['.'] if '.' in acd.keys() else 0)
    total_sites = sum(acd.values())-missing_inds
    return acd,total_sites,missing_inds



def isInvariant(rec):
    alleles, total_sites, missing_inds = getAlleleStats(rec)
    if len(alleles) == 1:
        return True
    return False

def isInformative(rec, mincount=2, alleles=None):
    count = 0
    if alleles is None:
        alleles, total_sites, missing_inds = getAlleleStats(rec)
    if len(alleles) != 2:
        return False
    i1,i2 = alleles.keys()
    return (alleles[i1] >= mincount and alleles[i2] >= mincount)

def getPassSites(record_list, remove_cpg=False, remove_indels=True,
                 remove_multiallele=True, remove_missing=0,
                 inform_level=2, fasta_ref=None):
    pass_list = [True for r in record_list]
    if remove_cpg == True and fasta_ref is None:
        raise Exception("CpG removal requires a reference")
    if inform_level > 2 or inform_level < 0:
        raise Exception("Inform level %d must be between 0 and 2" % inform_level)
    if remove_multiallele:
        checkForMultiallele(record_list,pass_list)
    for i in range(len(record_list)):
        rec = record_list[i]
        if remove_indels and not checkRecordIsSnp(rec):
            pass_list[i] = False
        if remove_cpg and checkIfCpG(rec,fasta_ref):
            pass_list[i] = False
        alleles,total_sites,missing_inds = getAlleleStats(rec)
        if remove_missing != -1 and missing_inds > int(remove_missing):
            pass_list[i] = False
        if inform_level != 0 and not isInformative(rec,mincount=inform_level,alleles=alleles):
            pass_list[i] = False
    #pp = zip([r.pos for r in record_list],pass_list)
    #for ppp in pp:
    #    logging.info(ppp)
    return pass_list

def checkRecordPass(rec, remove_cpg=False, remove_indels=True, 
                    remove_multiallele=True, remove_missing=0,
                    inform_level=2,fasta_ref=None):
    if remove_cpg and fasta_ref is None:
        raise Exception("CpG removal requires a reference")
    if inform_level > 2 or inform_level < 0:
        raise Exception("Inform level %d must be between 0 and 2" % inform_level)
    if remove_indels and not checkRecordIsSnp(rec):
        return False
    if remove_cpg and checkIfCpG(rec,fasta_ref):
        return False
    if remove_missing != -1 or inform_level != 0:
        alleles,total_sites,missing_inds = getAlleleStats(rec)
        if remove_missing != -1 and missing_inds > int(remove_missing):
            return False
        if inform_level != 0 and not isInformative(rec,mincount=inform_level,
                alleles=alleles):
            return False
    return True


def filterSites(record_list, remove_cpg=False, remove_indels=True,
                remove_multiallele=True, remove_missing=0, inform_level=2,
                fasta_ref=None):
    pass_list = getPassSites(record_list,remove_cpg,remove_indels,remove_multiallele,remove_missing,inform_level,fasta_ref)
    out_list = []
    for i in range(len(pass_list)):
        if pass_list[i]:
            out_list.append(record_list[i])
    return out_list

def crossModelAndVcf(pop_list,vcf_samples,allow_missing_inds=True):
    missing_list = []
    present_list = []
    for p in pop_list:
        if p not in vcf_samples:
            missing_list.append(p)
        else:
            present_list.append(p)
    if len(missing_list) != 0:
        if allow_missing_inds:
            logging.warning("Samples %s are missing from VCF" % (','.join(missing_list)))
        else:
            raise Exception("Samples %s are missing from VCF" % (','.join(missing_list)))
    return present_list

def getIndsWithAllData(rec_list):
    idx_list = []
    for i in range(len(rec_list[0].samples)):
        has_data = True
        for rec in rec_list:
            if rec.samples[i].alleles[0] in [None,'N']:
                has_data = False
        if has_data:
            idx_list.append(i)
    return idx_list


class VcfReader():
    def __init__(self, vcfname, compress_flag=False, subsamp_num=None,
                 subsamp_fn=None, subsamp_list=None, index=None, 
                 popmodel=None, use_allpop=False, allow_missing_inds=True):

        ext = checkFormat(vcfname)
        if ext in ['gzip','other'] :
            raise Exception(('Input file %s is gzip-formatted, must be either '
                             'uncompressed or zipped with bgzip' % vcfname))
        self.file_uncompressed = (ext == 'vcf')
        self.reader_uncompressed = (self.file_uncompressed and not compress_flag)
        self.popmodel = None
        self.popkeys = None
        if popmodel is not None and use_allpop:
            raise Exception("Popmodel and allpop cannot both be specified")
        if compress_flag and self.file_uncompressed:
            vcfname = compressVcf(vcfname)
        subsamp_list = None
        if subsamp_num is not None:
            if subsamp_list is not None:
                raise Exception('Multiple subsampling options called in getVcfReader')
            subsamp_list = getSubsampleList(vcfname, subsamp_num)
        elif subsamp_fn is not None:
            if subsamp_list is not None:
                raise Exception('Multiple subsampling options called in getVcfReader')
            subsamp_file = open(subsamp_fn,'r')
            subsamp_list = [l.strip() for l in subsamp_file.readlines()]
            subsamp_file.close()

        self.openSetInds(vcfname,index,popmodel,use_allpop,subsamp_list)
        self.info_rec = next(self.reader)
        self.prev_last_rec = None #next(self.reader)
        self.reader.close()
        self.openSetInds(vcfname,index,popmodel,use_allpop,subsamp_list)
        self.chr_in_chrom = (self.info_rec.chrom[0:3] == 'chr')

    def openSetInds(self, vcfname, index, popmodel, use_allpop, subsamp_list):
        if index is None:
            self.reader = pysam.VariantFile(vcfname)
        else:
            self.reader = pysam.VariantFile(vcfname, index_filename=index)
        if popmodel is not None:
            self.popmodel = popmodel
            popsamp_list = popmodel.inds
            vcf_names = [l for l in self.reader.header.samples]
            present_list = crossModelAndVcf(popsamp_list,vcf_names)
            #self.reader.subset_samples(popsamp_list)
            self.reader.subset_samples(present_list)
            self.setPopIdx(present_list)
        if use_allpop:
            self.setAllPop()
        if subsamp_list is not None:
            logging.debug('Subsampling %d individuals from VCF file' %
            (len(subsamp_list)))
            self.reader.subset_samples(subsamp_list)

    def fetch(self, chrom=None, start=None, end=None):
        return self.reader.fetch(chrom, start, end)

    def getRecordList(self, region=None, chrom=None, start=None,
                      end=None):
        if self.reader_uncompressed:
            ret, self.prev_last_rec = getRecordListUnzipped(self.reader, self.prev_last_rec, region, add_chr=self.chr_in_chrom)
            return ret
        else:
            return getRecordList(self.reader, region, chrom, start, end, self.chr_in_chrom)

    def setPopIdx(self,present_list):
        self.popkeys = {}
        sample_names = [l for l in self.reader.header.samples]
        for p in self.popmodel.pop_list:
            self.popkeys[p] = []
            for indiv in self.popmodel.ind_dict[p]:
                if indiv in present_list:
                    self.popkeys[p].append(sample_names.index(indiv))
                else:
                    self.popkeys[p].append(-1)

    def close(self):
        self.reader.close()

    def setAllPop(self):
        self.popkeys = {'ALL':[]}
        for i in range(len(self.reader.header.samples)):
            self.popkeys['ALL'].append(i)

    def returnNames(self,index_list):
        return [self.reader.header[i] for i in index_list]

    def getRegionIterUnzipped(self,region=None,add_chr=False):
        if region is None:
            a = self.getNext()
            while a is not None:
                #self.prev_last_rec = a
                yield a
                self.prev_last_rec = a
                a = self.getNext()
            self.prev_last_rec = None
            return
        trec = (self.prev_last_rec if self.prev_last_rec is not None else self.getNext())
        if trec is None:
            self.prev_last_rec = trec
            return

        self.prev_last_rec = trec
        stat = region.containsRecord(trec)
        if stat == 'after':
            return
        if stat == 'in':
            yield trec
            self.prev_last_rec = trec
            trec = self.getNext()
            if trec is None:
                self.prev_last_rec = None
                return
            stat = region.containsRecord(trec)
        elif stat == 'before':
            while stat == 'before':
                self.prev_last_rec = trec
                trec = self.getNext()
                if trec is None:
                    self.prev_last_rec = trec
                    return
                stat = region.containsRecord(trec)
        while stat != 'after':
            yield trec
            self.prev_last_rec = trec
            trec = self.getNext()
            if trec is None:
                self.prev_last_rec = trec
                return
            stat = region.containsRecord(trec)
        self.prev_last_rec = trec

        
    def getRegionIterZipped(self,region=None):
        if region is None:
            treader = self.reader.fetch()
        else:
            treader = self.reader.fetch(region.chrom,region.start,region.end)
        for rec in treader:
            yield rec
        return

    def getRegionIter(self,region=None):
        if self.reader_uncompressed:
            return self.getRegionIterUnzipped(region=region)
        else:
            return self.getRegionIterZipped(region=region)
        
    def getNext(self,set_prev=True):
        try:
            trec = next(self.reader)
            return trec
        except StopIteration as e:
            return None
            

def modChrom(c,vcf_chr):
    if c is None:
        return None
    if vcf_chr and c[:3] != 'chr':
        return 'chr'+c
    if not vcf_chr and c[:3] == 'chr':
        return c[3:]
    return c

def getRecordList(vcf_reader, region=None, chrom=None, start=None,
                  end=None, add_chr=False):
    """Returns list for use in subsampling from input file"""
    if region is not None:
        c = modChrom(region.chrom,add_chr)
        var_sites = vcf_reader.fetch(c, region.start, region.end)
    else:
        c = modChrom(chrom,add_chr)
        var_sites = vcf_reader.fetch(c, start, end)
    lst = []
    for rec in var_sites:
        lst.append(rec)
    return lst

    


def getRecordListUnzipped(vcf_reader, prev_last_rec, region=None, chrom=None,
                          start=None, end=None, add_chr=False):
    """Method for getting record list from unzipped VCF file.

    This method will sequentially look through a VCF file until it finds
    the given `start` position on `chrom`, then add all records to a list
    until the `end` position has been reached. Note that `prev_last_rec`
    must be kept track of externally to ensure that if consecutive regions
    are called, the record of the first variant outside the first region
    is not lost.

    Parameters
    ----------
    vcf_reader : pysam VariantFile object
        VCF reader initialized from other function
    region : region object
        Region object with start and end coordinates of region of interest.
    prev_last_rec : VariantRecord object
        Variable with last record read from VcfReader. Stored here so that
        if two adjacent regions are called, the overflow record from the
        first region is still included in the next region

    Returns
    -------
    lst : list
        List of records in given gene region
    prev_last_rec : VariantRecord object
        First record after target region, for use in next call

    """
    lst = []
    if region is None:
        #lst.append(prev_last_rec)
        for rec in vcf_reader:
            lst.append(rec)
        return lst, lst[-1]

    if (prev_last_rec is not None and
        region.containsRecord(prev_last_rec) == 'in'):
        lst.append(prev_last_rec)
    elif (prev_last_rec is not None and
         region.containsRecord(prev_last_rec) == 'after'):
        return []
    rec = next(vcf_reader,None)
    if rec is None:
        return lst,None
    place = region.containsRecord(rec)
    while rec is not None and place != 'after':
        if place == 'in':
            lst.append(rec)
        rec = next(vcf_reader,None)
        if rec is None:
            break
        place = region.containsRecord(rec)
    prev_last_rec = rec
    return lst, prev_last_rec


def checkRecordIsSnp(rec):
    """Checks if this record is a single nucleotide variant, returns bool."""
    if len(rec.ref) != 1:
        return False
    if rec.alts is None:
        return False
    for allele in rec.alts:
        if len(allele) != 1:
            return False
    return True


def getSubsampleList(vcfname, ss_count):
    """Returns a list of the first `ss_count` individuals in `vcfname`

    """

    vcf_o = pysam.VariantFile(vcfname)
    rec = next(vcf_o)
    vcf_o.close()
    lst = []
    for samp in rec.samples:
        lst.append(samp)
    return lst[:int(ss_count)]


def compressVcf(vcfname,forceflag=False,remove=False):
    """Runs bgzip and tabix on input VCF file.

    Using the pysam library, this function runs the bgzip and tabix utilities
    on the given input file. By default, this will not overwrite an existing
    zipped file, but will overwrite an existing index. `remove` can be set to
    delete the unzipped file.

    Parameters
    ----------
    vcfname : str
        Name of uncompressed VCF file
    forceflag : bool (False)
        If true, will overwrite (vcfname).gz if it exists
    remove : bool (False)
        If true, will delete uncompressed source file

    Returns
    -------
    cvcfname : str
        Filepath to compressed VCF file
    """
    cvcfname = vcfname+".gz"
    pysam.tabix_compress(vcfname,cvcfname,force=forceflag)
    pysam.tabix_index(cvcfname,preset="vcf",force=True)
    if remove:
        os.remove(vcfname)
    return cvcfname

def vcfRegionName(prefix, region, ext, oneidx=False,
                  halfopen=True, sep='-'):
    chrom = region.toStr(halfopen, oneidx, sep)
    return prefix+'_'+chrom+'.'+ext

def getRecordsInRegion(region, record_list):
    sub_list = []
    for i in range(len(record_list)):
        loc = region.containsRecord(record_list[i])
        if loc == "in":
            sub_list.append(record_list[i])
        elif loc == "after":
            break
    return sub_list





#def getVcfReader(args):
def getVcfReader(vcfname, compress_flag=False, subsamp_num=None,
                 subsamp_fn=None, subsamp_list=None, index=None):
    """Returns a reader for a given input VCF file.

    Given a filename, filetype, compression option, and optional Subsampling
    options, will return a pysam.VariantFile object for iteration and
    a flag as to whether this file is compressed or uncompressed.

    Parameters
    ----------
    vcfname : str
        Filename for VCF file. The extension of this file will be used to
        determine whether it is compressed or not unless `var_ext` is set.
    var_ext : str (None)
        Extension for VCF file if it is not included in the filename.
    compress_flag : bool (False)
        If filetype is uncompressed and this is set to true, will run
        compressVcf function.
    subsamp_num : int (None)
        If set, will randomly select `subsamp_num` individuals (not
        genotypes) from the input VCF file and return a reader with
        only those data.
    subsamp_fn : str (None)
        If set, will return a reader with only data from the samples listed
        in the file provided. Cannot be used with other subsampling options.
    subsamp_list : list (None)
        If set, will return reader with records containing only
        individuals named in the list. Cannot be used with other subsampling
        options.

    Returns
    -------
    vcf_reader : pysam.VariantFile
        A reader that can be iterated through for variant records. If
        compressed, it will be able to use the pysam fetch method, otherwise
        it must be read through sequentially
    reader_uncompressed : bool
        If True, VCF reader is uncompressed. This means the fetch method
        cannot be used and region access must be done using the
        "getRecordListUnzipped" method.

    """
    ext = checkFormat(vcfname)
    if ext in ['gzip','other'] :
        raise Exception(('Input file %s is gzip-formatted, must be either '
                         'uncompressed or zipped with bgzip' % vcfname))
    file_uncompressed = (ext == 'vcf')
    reader_uncompressed = (file_uncompressed and not compress_flag)
    if compress_flag and file_uncompressed:
        vcfname = compressVcf(vcfname)
    #subsamp_list = None
    if subsamp_num is not None:
        if subsamp_list is not None:
            raise Exception('Multiple subsampling options called in getVcfReader')
        subsamp_list = getSubsampleList(vcfname, subsamp_num)
    elif subsamp_fn is not None:
        if subsamp_list is not None:
            raise Exception('Multiple subsampling options called in getVcfReader')
        subsamp_file = open(subsamp_fn,'r')
        subsamp_list = [l.strip() for l in subsamp_file.readlines()]
        subsamp_file.close()
    if index is None:
        vcf_reader = pysam.VariantFile(vcfname)
    else:
        vcf_reader = pysam.VariantFile(vcfname, index_filename=index)
    if subsamp_list is not None:
        logging.debug('Subsampling %d individuals from VCF file' %
        (len(subsamp_list)))
        vcf_reader.subset_samples(subsamp_list)
    return vcf_reader, reader_uncompressed



#!/usr/bin/env python
'''
    Automates the calculation of multiple admixture statistics, including: Patterson's D, 
    F4 statistic, F4-ratio statistic, and F3 statistic.

    ############################
    Input Command-line Arguments
    ############################
    **--eigenstrat-prefix** *<input_prefix>*
        Argument used to define the filename prefix shared by the genotype file (.geno), 
        the individual file (.ind), and the SNP file (.snp). Should not be used alongside
        the specific file arguments (e.g. --geno).
    **--geno** *<geno_filename>*
        Argument used to define the filename of the eigenstrat genotype file (.geno). 
        Must be called alongside --ind and --snp. Cannot be called alongside 
        --eigenstrat-prefix.
    **--ind** *<ind_filename>*
        Argument used to define the filename of the eigenstrat individual file (.ind). 
        Must be called alongside --geno and --snp. Cannot be called alongside 
        --eigenstrat-prefix.
    **--snp** *<snp_filename>*
        Argument used to define the filename of the eigenstrat SNP file (.snp). Must be
        called alongside --geno and --ind. Cannot be called alongside --eigenstrat-prefix.
    **--model-file** *<model_filename>*
        Argument used to define the model file. Please note that this argument cannot be 
        used with the individual-based filters.
    **--model** *<model_str>*
        Argument used to define the model (i.e. the individual(s) to include and/or the 
        populations for relevant statistics). May be used with any statistic. Please note 
        that this argument cannot be used with **--pop-file** argument or the 
        individual-based filters.
    
    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
        Cannot be used if multiple output files are created.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ####################################
    Statistic Command-line Specification
    ####################################
    **--calc-admix-statistic** *<D, F4, F4-ratio, F3>*
        Argument used to define the admix statistic to be calculated. Patterson's D (D), 
        F4 statistic (F4), F4-ratio statistic (F4-ratio), and F3 statistic (F3). See below
        for details on the arguments requried by each statistic .

    ***********************************
    Statistic Command-line Requirements
    ***********************************
    It should be noted that each admix statistic has a specific set of population labels
    arguments. These labels are used to specify a representive population. For instance, 
    the argument '--admix-w-pop CEU' will replace the W label of Patterson's D and the 
    F4 statistic with the CEU population. These arguments may be found in the next section.
 
    **--calc-admix-statistic** *D*
        Requires: **--admix-w-pop**/**--admix-w-pop-file**, 
        **--admix-x-pop**/**--admix-x-pop-file**,
        **--admix-y-pop**/**--admix-y-pop-file**, 
        and **--admix-z-pop**/**--admix-z-pop-file**.

    **--calc-admix-statistic** *F4*
        Requires: **--admix-w-pop**/**--admix-w-pop-file**, 
        **--admix-x-pop**/**--admix-x-pop-file**,
        **--admix-y-pop**/**--admix-y-pop-file**, 
        and **--admix-z-pop**/**--admix-z-pop-file**.

    **--calc-admix-statistic** *F4-ratio*
        Requires: **--admix-a-pop**/**--admix-a-pop-file**,
        **--admix-b-pop**/**--admix-b-pop-file**,
        **--admix-c-pop**/**--admix-c-pop-file**, 
        **--admix-x-pop**/**--admix-x-pop-file**,
        and **--admix-o-pop**/**--admix-o-pop-file**.
    **--calc-admix-statistic** *F3*
        Requires: **--admix-a-pop**/**--admix-a-pop-file**,
        **--admix-b-pop**/**--admix-b-pop-file**,
        and **--admix-c-pop**/**--admix-c-pop-file**.

    *******************************************
    Additional Statistic Command-line Arguments
    *******************************************
    **--admix-w-pop** *<w_pop_str>* *<w_pop1_str, w_pop2_str, etc.>*
        Argument used to define the population(s) to represent W in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented W.
    **--admix-w-pop-file** *<w_pop_filename>*
        Argument used to define a file of population(s) to represent W in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented W.
    **--admix-x-pop** *<x_pop_str>* *<x_pop1_str, x_pop2_str, etc.>*
        Argument used to define the population(s) to represent X in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented X.
    **--admix-x-pop-file** *<x_pop_filename>*
        Argument used to define a file of population(s) to represent X in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented X.
    **--admix-y-pop** *<y_pop_str>* *<y_pop1_str, y_pop2_str, etc.>*
        Argument used to define the population(s) to represent Y in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented Y.
    **--admix-y-pop-file** *<y_pop_filename>*
        Argument used to define a file of population(s) to represent Y in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented Y.
    **--admix-z-pop** *<z_pop_str>* *<z_pop1_str, z_pop2_str, etc.>*
        Argument used to define the population(s) to represent Z in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented Z.
    **--admix-z-pop-file** *<z_pop_filename>*
        Argument used to define a file of population(s) to represent Z in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented Z.
    **--admix-a-pop** *<a_pop_str>* *<a_pop1_str, a_pop2_str, etc.>*
        Argument used to define the population(s) to represent A in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented A.
    **--admix-a-pop-file** *<a_pop_filename>*
        Argument used to define a file of population(s) to represent A in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented A.
    **--admix-b-pop** *<b_pop_str>* *<b_pop1_str, b_pop2_str, etc.>*
        Argument used to define the population(s) to represent B in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented B.
    **--admix-b-pop-file** *<b_pop_filename>*
        Argument used to define a file of population(s) to represent B in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented B.
    **--admix-c-pop** *<c_pop_str>* *<c_pop1_str, c_pop2_str, etc.>*
        Argument used to define the population(s) to represent C in the supported admixure 
        statistic. This argument may be used multiple times if desired. If multiple 
        populations the statistic will be repeated until each population has represented C.
    **--admix-c-pop-file** *<c_pop_filename>*
        Argument used to define a file of population(s) to represent C in the supported 
        admixure statistic. If multiple populations the statistic will be repeated until each 
        population has represented C.
    
    ##########################
    Example Command-line Usage
    ##########################
    Command-line to calculate Patterson's D:

    .. code-block:: bash
        
        python eigenstrat_fstats.py --eigenstrat-prefix snps --calc-admix-statistic D --admix-w-pop French --admix-x-pop Yoruba --admix-y-pop Vindija --admix-z-pop Chimp 
    
    Command-line to calculate the F4-ratio: 

    .. code-block:: bash
        
        python eigenstrat_fstats.py --eigenstrat-prefix snps --calc-admix-statistic F4-ratio --admix-a-pop Altai --admix-b-pop Vindija --admix-c-pop Yoruba --admix-x-pop French --admix-o-pop Chimp  
'''

import os
import sys
import argparse
import logging
import subprocess
import random
import string
import shutil

# Import PPP modules and scripts
from pgpipe.admixtools import *
from pgpipe.model import read_model_file, pops_not_in_model
from pgpipe.logging_module import initLogger, logArgs

def admix_parser (passed_arguments):
    '''admix Argument Parser - Assigns arguments from command line'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]

                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    admix_parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Input Eigenstrat prefix arguments
    admix_parser.add_argument("--eigenstrat-prefix", help = "Defines the filename prefix of the GENO, IND, and SNP files", type = str)

    # Input Eigenstrat file arguments
    admix_parser.add_argument("--geno", dest = 'geno_filename', help = "Defines the filename of the GENO file. Called alongside --ind and --snp", type = str, action = parser_confirm_file())
    admix_parser.add_argument("--ind", dest = 'ind_filename', help = "Defines the filename of the IND file. Called alongside --geno and --snp", type = str, action = parser_confirm_file())
    admix_parser.add_argument("--snp", dest = 'snp_filename', help = "Defines the filename of the SNP file. Called alongside --geno and --ind", type = str, action = parser_confirm_file())

    # Model file arguments.
    admix_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    admix_parser.add_argument('--model', help = 'Defines the model and the individual(s) to include', type = str)

    admix_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', default = 'out')

    # General arguments.
    admix_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')

    # Statistic based arguments
    stat_list = ['D', 'F4', 'F4-ratio', 'F3']
    admix_parser.add_argument('--calc-admix-statistic', metavar = metavar_list(stat_list), help = 'The admixture statistic to calculate', required = True, type = str, choices = stat_list)
    
    # Admix analysis pops
    admix_parser.add_argument('--admix-w-pop', dest = 'w_pops', help = 'W population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-w-pop-file', dest = 'w_file', help = 'W populations for admixure analysis', type = str, action = parser_confirm_file())
    
    admix_parser.add_argument('--admix-x-pop', dest = 'x_pops', help = 'X population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-x-pop-file', dest = 'x_file', help = 'X populations for admixure analysis', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-y-pop', dest = 'y_pops', help = 'Y population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-y-pop-file', dest = 'y_file', help = 'Y populations for admixure analysis', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-z-pop', dest = 'z_pops', help = 'Z population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-z-pop-file', dest = 'z_file', help = 'Z populations for admixure analysiss', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-a-pop', dest = 'a_pops', help = 'A population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-a-pop-file', dest = 'a_file', help = 'A populations for admixure analysis', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-b-pop', dest = 'b_pops', help = 'B population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-b-pop-file', dest = 'b_file', help = 'B populations for admixure analysis', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-c-pop', dest = 'c_pops', help = 'C population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-c-pop-file', dest = 'c_file', help = 'C populations for admixure analysis', type = str, action = parser_confirm_file())

    admix_parser.add_argument('--admix-o-pop', dest = 'o_pops', help = 'O population(s) for admixure analysis', nargs = '+', type = str, action = parser_add_to_list())
    admix_parser.add_argument('--admix-o-pop-file', dest = 'o_file', help = 'O populations for admixure analysis', type = str, action = parser_confirm_file())

    if passed_arguments:
        return admix_parser.parse_args(passed_arguments)
    else:
        return admix_parser.parse_args()

def read_admix_pops_file (pop_filename):

    # List to hold admix pops 
    admix_pops = []

    # Open the ind file
    with open(pop_filename, 'r') as pop_file:

        # Loop the ind file
        for pop_line in pop_file:

            # Append the population
            admix_pops.append(pop_line.strip())

    # Return the admix pops
    return admix_pops

def run(passed_arguments = []):

    # Grab admixtools arguments from command line
    admix_args = admix_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(admix_args, func_name='admixtools')

    # Read in the models
    models_in_file = read_model_file(admix_args.model_file)

    # Check that the selected model was not found in the file
    if admix_args.model not in models_in_file:
        raise IOError('Selected model "%s" not found in: %s' % (admix_args.model, admix_args.model_file))

    # Select model, might change this in future versions
    selected_model = models_in_file[admix_args.model]
      
    # Check if the W populations file has been assigned
    if admix_args.w_file:

        # Save the pops from the file
        w_file_pops = read_admix_pops_file(admix_args.w_file)

        # Extend the pops list
        admix_args.w_pops.extend(w_file_pops)

    # Check if the X populations file has been assigned
    if admix_args.x_file:

        # Save the pops from the file
        x_file_pops = read_admix_pops_file(admix_args.x_file)

        # Extend the pops list
        admix_args.x_pops.extend(x_file_pops)

    # Check if the Y populations file has been assigned
    if admix_args.y_file:

        # Save the pops from the file
        y_file_pops = read_admix_pops_file(admix_args.y_file)

        # Extend the pops list
        admix_args.y_pops.extend(y_file_pops)

    # Check if the Z populations file has been assigned
    if admix_args.z_file:

        # Save the pops from the file
        z_file_pops = read_admix_pops_file(admix_args.z_file)

        # Extend the pops list
        admix_args.z_pops.extend(z_file_pops)

    # Check if the A populations file has been assigned
    if admix_args.a_file:

        # Save the pops from the file
        a_file_pops = read_admix_pops_file(admix_args.a_file)

        # Extend the pops list
        admix_args.a_pops.extend(a_file_pops)

    # Check if the B populations file has been assigned
    if admix_args.b_file:

        # Save the pops from the file
        b_file_pops = read_admix_pops_file(admix_args.b_file)

        # Extend the pops list
        admix_args.b_pops.extend(b_file_pops)

    # Check if the C populations file has been assigned
    if admix_args.c_file:

        # Save the pops from the file
        c_file_pops = read_admix_pops_file(admix_args.c_file)

        # Extend the pops list
        admix_args.c_pops.extend(c_file_pops)

    # Check if the O populations file has been assigned
    if admix_args.o_file:

        # Save the pops from the file
        o_file_pops = read_admix_pops_file(admix_args.o_file)

        # Extend the pops list
        admix_args.o_pops.extend(o_file_pops)

    # Check if W pop(s) were assigned
    if admix_args.w_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.w_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('W population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if X pop(s) were assigned
    if admix_args.x_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.x_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('X population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if Y pop(s) were assigned
    if admix_args.y_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.y_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('Y population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if Z pop(s) were assigned
    if admix_args.z_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.z_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('Z population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if A pop(s) were assigned
    if admix_args.a_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.a_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('A population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if B pop(s) were assigned
    if admix_args.b_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.b_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('B population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if C pop(s) were assigned
    if admix_args.c_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.c_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('C population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Check if O pop(s) were assigned
    if admix_args.o_pops:

        # Assign pops not found within the model
        pops_not_found = pops_not_in_model(selected_model, admix_args.o_pops)

        # Check if any pops were not found
        if pops_not_found:
            raise Exception('O population(s) not found in model: %s' % ', '.join(pops_not_found))

    # Bool to indicate if file names were changed
    eigenstrat_filenames_changed = False

    # Check if an eigenstrat prefix was assigned
    if admix_args.eigenstrat_prefix:

        # Confirm all eigenstrat files from prefix, will raise exception upon failure
        confirm_eigenstrat_files_from_prefix(admix_args.eigenstrat_prefix)

    # Check if the eigenstrat files were assigned separately
    elif admix_args.geno_filename or admix_args.ind_filename or admix_args.snp_filename:

        # Confirm that all eigenstrat files exist, will raise exception upon failure
        confirm_eigenstrat_files(admix_args.geno_filename, admix_args.ind_filename, admix_args.snp_filename)

        # Confirm that an eigenstrat prefix may be assigned
        if check_eigenstrat_prefix(admix_args.geno_filename, admix_args.ind_filename, admix_args.snp_filename):

            # Assign the prefix
            admix_args.eigenstrat_prefix = os.path.splitext(admix_args.geno_filename)[0]

        else:

            # Confirm the eigenstrat filenames were changed
            eigenstrat_filenames_changed = True

            # Rename files and assign the prefix
            admix_args.eigenstrat_prefix = assign_eigenstrat_prefix(admix_args.geno_filename, admix_args.ind_filename, admix_args.snp_filename)

    # Assign a ind filename with the model
    model_ind_filename = assign_model_ind_filename(admix_args.eigenstrat_prefix, selected_model, overwrite = admix_args.overwrite)
    
    # Create a population filename
    create_ind_w_pops(admix_args.eigenstrat_prefix, selected_model, model_ind_filename)

    # String to hold the expected statistic output
    expected_statistic_output = ''

    # Check if the D statistic was specified
    if admix_args.calc_admix_statistic == 'D':

        # Check that the all the requried pops were specified
        if not admix_args.w_pops or not admix_args.x_pops or not admix_args.y_pops or not admix_args.z_pops:
            raise Exception('D requires the the following populations to be defined: W, X, Y, and Z')

        # Assign the expected statistic output 
        expected_statistic_output = admix_args.out_prefix + '.d'

        # Run the D Statistic
        r_admixr_d(admix_args.eigenstrat_prefix, admix_args.out_prefix, admix_args.w_pops, admix_args.x_pops, admix_args.y_pops, admix_args.z_pops)

    # Check if the F4 statistic was specified
    elif admix_args.calc_admix_statistic == 'F4':

        # Check that the all the requried pops were specified
        if not admix_args.w_pops or not admix_args.x_pops or not admix_args.y_pops or not admix_args.z_pops:
            raise Exception('F4 requires the the following populations to be defined: W, X, Y, and Z')

        # Assign the expected statistic output
        expected_statistic_output = admix_args.out_prefix + '.f4'
        
        # Run the F4 Statistic
        r_admixr_f4(admix_args.eigenstrat_prefix, admix_args.out_prefix, admix_args.w_pops, admix_args.x_pops, admix_args.y_pops, admix_args.z_pops)

    elif admix_args.calc_admix_statistic == 'F4-ratio':

        # Check that the all the requried pops were specified
        if not admix_args.a_pops or not admix_args.b_pops or not admix_args.c_pops or not admix_args.x_pops or not admix_args.o_pops:
            raise Exception('F4-ratio requires the the following populations to be defined: A, B, C, X, and O')

        # Assign the expected statistic output
        expected_statistic_output = admix_args.out_prefix + '.f4ratio'

        # Run the F4ratio Statistic
        r_admixr_f4ratio(admix_args.eigenstrat_prefix, admix_args.out_prefix, admix_args.a_pops, admix_args.b_pops, admix_args.c_pops, admix_args.x_pops, admix_args.o_pops)

    elif admix_args.calc_admix_statistic == 'F3':

        # Check that the all the requried pops were specified
        if not admix_args.a_pops or not admix_args.b_pops or not admix_args.c_pops:
            raise Exception('F3 requires the the following populations to be defined: A, B, and C')

        # Assign the expected statistic output
        expected_statistic_output = admix_args.out_prefix + '.f3'

        # Run the F3 Statistic
        r_admixr_f3(admix_args.eigenstrat_prefix, admix_args.out_prefix, admix_args.a_pops, admix_args.b_pops, admix_args.c_pops)

    # Remove the model ind file
    os.remove(model_ind_filename)

    # Check if the eigenstrat filenames were changed
    if eigenstrat_filenames_changed:

        # Restore the filename if they were changed
        restore_eigenstrat_files(admix_args.geno_filename, admix_args.ind_filename, admix_args.snp_filename)

if __name__ == "__main__":
    initLogger()
    run()



"""
    reads an ima3 input file
    generates a gzipped treemix file
    takes two arguments at the command line
        input file
        output file

    the output file is to be run using treemix's block option
    the data from a locus in the ima3 file is padded out to have a length of
    200 snps  (i.e. padded with uninformative snps)

    Then treemix is run with -k 200

"""
import sys
import gzip

def getVariantSites(seq_list):
    idx = []
    for i in range(len(seq_list[0])):
        first_base = seq_list[0][i]
        for j in range(1,len(seq_list)):
            cur_base = seq_list[j][i]
            if first_base != cur_base:
                idx.append(i)
                break
    return idx



def parseSeqs(seq_list, seq_count, padline,padcount):
    idx = getVariantSites(seq_list)
    var_list = []
    for i in idx:
        ref_count = [0 for i in range(len(seq_count))]
        alt_count = [0 for i in range(len(seq_count))]
        ref = seq_list[0][i]
        #ref_count[0] = 1
        cur_seq = 0
        for j in range(len(seq_count)):
            for k in range(seq_count[j]):
                if seq_list[cur_seq][i] == ref:
                    ref_count[j] += 1
                else:
                    alt_count[j] += 1
                cur_seq += 1
        var_line = ' '.join([str(a)+','+str(b) for a,b in zip(ref_count,alt_count)])
        var_list.append(var_line)
    if padcount < len(var_list):
        raise Exception("too many snps")
    while len(var_list) < padcount:
        var_list.append(padline)
    return var_list

def main(imfilename,treemixfilename):

    padcount = 200
    f = open(str(imfilename),'r')
    locus_list = []

    pop_list = []

    output_contents = []

    l = f.readline()
    l = f.readline()
    while l[0] == '#':
        l = f.readline()
    npops = l.split()[0]
    pop_list = f.readline().strip().split()
    print ("populations",pop_list)
    output_contents.append(' '.join(pop_list))
    temp = f.readline().strip()
    print ("number of loci:",temp)
    if temp.find('(') >= 0:
        treestring = temp
        num_loci = int(f.readline().strip())
    else:
        num_loci = int(temp)
    for i in range(num_loci):
        loci_data = f.readline().strip()
        loci_array = loci_data.split()
        seq_count = [int(ll) for ll in loci_array[1:1+len(pop_list)]]
        seq_total = sum(seq_count)
        padline = ""
        for c in seq_count:
            padline  = padline + str(c) + ",0 "
        seq_list = []
        for j in range(seq_total):
            seq_list.append(f.readline().strip().split()[1])
        sl = parseSeqs(seq_list, seq_count,padline[0:-1],padcount)
        for s in sl:
            output_contents.append(s)

    out_buffer = '\n'.join(output_contents)+'\n'
    if str(treemixfilename).endswith('gz'):
        with gzip.open(str(treemixfilename), 'wb') as out_f:
            out_f.write(out_buffer.encode())
    else:
        with gzip.open(str(treemixfilename)+".gz", 'wb') as out_f:
            out_f.write(out_buffer.encode())



if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("reads an ima3 input file\ngenerates a gzipped treemix file\ntakes two arguments at the command line\n\tinput file\toutput file")
    main(sys.argv[1],sys.argv[2])




import os
import sys
import argparse
import itertools
import logging
import shutil
import copy
import numpy as np
import pandas as pd
from collections import defaultdict

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

# Import log initializer
from pgpipe.logging_module import initLogger, logArgs

# Model file related functions
from pgpipe.model import read_model_file, write_model_file

def sampler_parser(passed_arguments):
    '''Sampler Argument Parser - Assigns arguments from command line.'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string = None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_dict_list ():
        '''Custom action to house data as defaultdict int'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):

                argument_call = self.dest.replace('_','-')

                if not isinstance(values, (list, tuple)):
                    raise Exception('--%s argument assignment error. Expected iterable' % argument_call)

                if len(values) != 2:
                    raise Exception('--%s argument assignment error. Expected iterable with a length 2' % argument_call)

                # Assign the passed value
                if getattr(args, self.dest):
                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]].append(values[1])

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(list)
                    arg_dict[values[0]].append(values[1])
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_int ():
        '''Custom action to house data as defaultdict int'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):

                argument_call = self.dest.replace('_','-')

                if not isinstance(values, (list, tuple)):
                    raise Exception('--%s argument assignment error. Expected iterable' % argument_call)

                if len(values) != 2:
                    raise Exception('--%s argument assignment error. Expected iterable with a length 2' % argument_call)

                try:
                    int(values[1])
                except:
                    raise Exception('--%s argument assignment error. Expected integer as second value' % argument_call)

                # Assign the passed value
                if getattr(args, self.dest):

                    if values[0] in getattr(args, self.dest):
                        raise Exception('--%s argument assignment error. Models may only be sampled once' % argument_call)

                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]] = int(values[1])

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(int)
                    arg_dict[values[0]] = int(values[1])
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_dict ():
        '''Custom action to house data as defaultdict dict'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None, split_char = '='):

                argument_call = self.dest.replace('_','-')

                if not isinstance(values, (list, tuple)):
                    raise Exception('--%s argument assignment error. Expected iterable' % argument_call)

                if len(values) < 2:
                    raise Exception('--%s argument assignment error. Expected iterable with length >= 2' % argument_call)

                # Create dict to hold the non-key values
                value_dict = {}

                # Loop the non-key values
                for sample_value in values[1:]:

                    # Check that the sample_value may be split using split_char
                    if split_char not in sample_value:
                        # Report an error if the data cannot be split
                        raise Exception('--%s argument assignment error. Populations could not be split' % argument_call)

                    # Split the value using the split_char argument
                    split_value = sample_value.split(split_char)
                    # Add the split value to value_dict
                    value_dict[split_value[0]] = split_value[1]

                # Assign the passed value
                if getattr(args, self.dest):

                    if values[0] in getattr(args, self.dest):
                        raise Exception('--%s argument assignment error. Models may only be sampled once' % argument_call)

                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]] = value_dict

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(dict)
                    arg_dict[values[0]] = value_dict
                    setattr(args, self.dest, arg_dict)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    sampler_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments
    sampler_parser.add_argument('--model-file', help = 'Model file to sample from', required = True, type = str, action = parser_confirm_file())

    # Sampler arguments
    sampler_parser.add_argument('--model-copy', help = 'Copy the specified model', type = str, nargs = 2, action = parser_dict_list())

    sampler_parser.add_argument('--model-sample-pops', help = 'Sample all populations within the selected model', type = str, nargs = 2, action = parser_dict_int())

    sampler_parser.add_argument('--model-sample-pop', help = 'Sample all populations within the selected model', type = str, nargs = '+', action = parser_dict_dict())

    # New-file options
    sampler_parser.add_argument('--out', help = 'Specifies the filename of the new model file', type = str, default = 'out.model')
    sampler_parser.add_argument('--keep-unsampled', help = 'Specifies if an output file should keep unsampled models', action = 'store_true')

    # Random options
    sampler_parser.add_argument('--random-seed', help="Defines the random seed value for the random number generator", type = int, default = np.random.randint(1, high = 1000000000))

    if passed_arguments:
        return sampler_parser.parse_args(passed_arguments)
    else:
        return sampler_parser.parse_args()

def sample_model_w_dict (model_file_data, model_to_sample, sample_pop_dict):

    # Check if the selected model was not found in the file
    if model_to_sample not in model_file_data:
        raise IOError('Selected model "%s" not found in %s' % (model_to_sample, model_file_data))

    # Loop the population sampleing dict
    for sample_pop_name, sample_pop_int in sample_pop_dict.items():

        # Sample the selected pop within the specified model
        model_file_data[model_to_sample].sample_pop(sample_pop_name, sample_pop_int)

def sample_model_w_int (model_file_data, model_to_sample, sample_pops_int):

    # Check if the selected model was not found in the file
    if model_to_sample not in model_file_data:
        raise IOError('Selected model "%s" not found in %s' % (model_to_sample, model_file_data))

    # Sample the selected model using the sample pops int
    model_file_data[model_to_sample].sample_pops(sample_pops_int)

def copy_model (model_file_data, model_to_copy, model_from_copy):

    # Check if the source model was not found in the file
    if model_to_copy not in model_file_data:
        raise IOError('Source model for copy "%s" not found in %s' % (model_to_copy, model_file_data))

    # Check if the copy is found in the file
    if model_from_copy in model_file_data:
        raise IOError('Copy "%s" already in %s' % (model_from_copy, model_file_data))

    # Copy the model
    model_file_data.copy_model(model_to_copy, model_from_copy)

def run (passed_arguments = []):
    '''
        Model sampler.

        Automates the sampling process of a specified model file. The function
        allows the user to sample populations to contain a specific number of
        individuals.

        Parameters
        ----------
        --model-file : str
            Model file to sample from
        --model-copy : str str
            Copy the specified model (i.e. source) as a new model (i.e. target).
            Example: --model-copy source_model_name target_model_name
        --model-sample-pops : str int
            Sample the population within the specified model using the given
            integer. Example: --model-copy model_name k
        --model-sample-pop : str str=int
            Sample the specified population within the specified model using the
            given integer. Example: --model-copy model_name pop=k
        --random-seed : int
            Specifies the random seed value for the random number generator
        --out : str
            Filename of the sampled model output
        --keep-unsampled :
            Specifies if unsampled models should be saved in the output

        Returns
        -------
        output : file
            Sampled statistic file
        log : file
            Log file output

        Raises
        ------
        IOError
            Model file does not exist
        Exception
            If user attepts to sample a model more than once
    '''

    # Get arguments from command line
    sampler_args = sampler_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(sampler_args, func_name = 'model_sampler')

    # Check that no model is found in both --model-sample-pop/--model-sample-pops
    if set(list(sampler_args.model_sample_pop)) & set(list(sampler_args.model_sample_pops)):
        raise Exception('Argument error. Sampling processes found in both --model-sample-pop and --model-sample-pops')

    # Set the random seed
    np.random.seed(sampler_args.random_seed)

    logging.info('Random seed: %s' % sampler_args.random_seed)

    # Read in the model file
    model_file_data = read_model_file(sampler_args.model_file)

    # Check if any models will be copied, should copy prior to sampling
    if sampler_args.model_copy:

        # Loop the copy entries
        for model_to_copy, models_from_copy in sampler_args.model_copy.items():

            # Loop the models created from the copy
            for model_from_copy in models_from_copy:

                # Copy the model
                copy_model(model_file_data, model_to_copy, model_from_copy)

            logging.info('%s successfully copied to %s' % (model_to_copy, models_from_copy))

    # Check if any models will sampled using the pops method
    if sampler_args.model_sample_pops:

        # Loop the entries
        for model_to_sample, sample_pops_int in sampler_args.model_sample_pops.items():

            sample_model_w_int(model_file_data, model_to_sample, sample_pops_int)

            logging.info('%s successfully sampled by %s' % (model_to_sample, sample_pops_int))

    # Check if any models will sampled using the pop method
    if sampler_args.model_sample_pop:

        # Loop the entries
        for model_to_sample, sample_pops_dict in sampler_args.model_sample_pop.items():

            sample_model_w_dict(model_file_data, model_to_sample, sample_pops_dict)

        logging.info('%s successfully sampled by each population' % model_to_sample )

    # Check if the sampled models should be deleted
    if not sampler_args.keep_unsampled:

        # Create a complete list of the sampled models
        sampled_models = list(sampler_args.model_sample_pop) + list(sampler_args.model_sample_pops)

        # Create lite of unsampled models
        unsampled_models =  list(set(model_file_data) - set(sampled_models))

        # Loop the unsampled models
        for unsampled_model in unsampled_models:

            # Delete the unsampled model
            del model_file_data[unsampled_model]

        logging.info('Unsampled model removed')

    # Create the model output file
    write_model_file(model_file_data, sampler_args.out)

if __name__ == "__main__":
    initLogger()
    run()



#!/usr/bin/env python
'''
    Automates the calculation of site/windowed fixation index (Fst), Tajima's D,
    nucleotide diversity (Pi), allele frequency, and heterozygosity using
    VCFTools. If no statistic is specified, windowed Fst is used by default.

    ##################
    Command-line Usage
    ##################
    The VCF statistic calculator may be called using the following command:

    .. code-block:: bash
        
        python vcf_calc.py

    *************
    Example usage
    *************
    Command-line to calculate Tajima's D:

    .. code-block:: bash
        
        python vcf_calc.py --vcf input.vcf --calc-statistic TajimaD --statistic-window-size 10000

    Command-line to calculate windowed Fst on the two populations within the model *2Pop*: 

    .. code-block:: bash
        
        python vcf_calc.py --vcf input.vcf.gz --model-file input.model --model 2Pop --calc-statistic windowed-weir-fst --statistic-window-size 10000 --statistic-window-step 10000
    
    ############
    Dependencies 
    ############
    * `VCFtools <https://vcftools.github.io/index.html>`_
     
    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file for calculations.
    **--model-file** *<model_filename>*
        Argument used to define the model file. Please note that this argument cannot be 
        used with the **--pop-file** argument or individual-based filters.
    **--model** *<model_str>*
        Argument used to define the model (i.e. the individual(s) to include and/or the 
        populations for relevant statistics). May be used with any statistic. Please note 
        that this argument cannot be used with **--pop-file** argument or the 
        individual-based filters.
    
    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
        Cannot be used if multiple output files are created.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-dir** *<output_dir_name>*
        Argument used to define the output directory. Only used if multiple output files 
        are created.
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ####################################
    Statistic Command-line Specification
    ####################################
    **--calc-statistic** *<weir-fst, windowed-weir-fst, TajimaD, site-pi, window-pi, freq, het-fit, het-fis, hardy-weinberg>*
        Argument used to define the statistic to be calculated. Site Fst (weir-fst), 
        windowed Fst (windowed-weir-fst), Tajima's D (TajimaD), site nucleotide 
        diversity (site-pi), windowed nucleotide diversity (window-pi), allele frequency
        (freq), Fit (het-fit), Fis (het-fis), and the hardy-weinberg equilibrium 
        (hardy-weinberg).

    ***********************************
    Statistic Command-line Requirements
    ***********************************
    It should be noted that some of the statistics in the VCF calculator require additional
    arguments (i.e. **--pop-file**, **--statistic-window-size**, **--statistic-window-step**).
    These statistics may be found below with their additional requirements. If a statistic is
    not given, only the statistic specification (i.e. **--calc-statistic**) is required.

    **--calc-statistic** *weir-fst*
        Requires: **--pop-file**/**--model**.

    **--calc-statistic** *windowed-weir-fst*
        Requires: **--pop-file**/**--model**, **--statistic-window-size**, and
        **--statistic-window-step**.

    **--calc-statistic** *TajimaD*
        Requires: **--statistic-window-size**

    **--calc-statistic** *windowed-pi*
        Requires: **--statistic-window-size** and **--statistic-window-step**.

    **--calc-statistic** *het-fis*
        Requires: **--pop-file**/**--model**.
    
    *******************************************
    Additional Statistic Command-line Arguments
    *******************************************
    **--statistic-window-size** *<size_int>*
        Defines the statistic window size. Not usable with all statistics.
    **--statistic-window-step** *<step_int>*
        Defines the statistic window step size. Not usable with all statistics.
    **--pop-file** *<pop_filename>*
        Population file. This argument may be used multiple times if desired. Please note
        the this argument is not compatible with either the **--model** or **--model-file** 
        command-line arguments.
    
    #############################
    Filter Command-line Arguments
    #############################
    If using an unfiltered VCF file (e.g. reduce the creation of unnecessary large files)
    the VCF calculator is able to use either a kept or removed sites/BED file and the
    individual-based paramemeters. 
    
    **************************
    Individual-Based Arguments
    **************************
    Please note that all individual-based arguments are not compatible with either the 
    **--model** or **--model-file** command-line arguments.

    **--filter-include-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to include. This argument may be used 
        multiple times if desired.
    **--filter-exclude-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to exclude. This argument may be used 
        multiple times if desired.
    **--filter-include-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to include.
    **--filter-exclude-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to exclude.
    
    ************************
    Position-Based Arguments
    ************************
    **--filter-include-positions** *<position_filename>*
        Argument used to define a file of positions to include within a tsv file 
        (chromosome and position).
    **--filter-exclude-positions** *<position_filename>*
        Argument used to define a file of positions to exclude within a tsv file 
        (chromosome and position).
    **--filter-include-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to include.
    **--filter-exclude-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to exclude.
'''

import os
import sys
import argparse
import itertools
import copy
import shutil
import logging

# Import basic vcftools functions
from pgpipe.vcftools import *

# Model file related functions
from pgpipe.model import read_model_file

# Import logging functions
from pgpipe.logging_module import initLogger, logArgs

def vcf_calc_parser(passed_arguments):
    '''
    VCF Calc Argument Parser

    Assign the parameters for the VCF Calc using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_confirm_file_list ():
        '''Custom action to confirm file exists in list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                # Loop the list
                for value_item in value:
                    # Check if the file exists
                    if not os.path.isfile(value_item):
                        raise IOError('%s not found' % value_item)
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]
                
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())

    # Model file arguments
    vcf_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    vcf_parser.add_argument('--model', help = 'Defines the model and the individual(s)/population(s) to include', type = str)

    # Non-model file arguments
    vcf_parser.add_argument('--pop-file', help = 'Population file. May be used multiple times', nargs = '+', type = str, action = parser_confirm_file_list())

    # Other file arguments. Expand as needed
    vcf_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix. Cannot be used if multiple output files are created', type = str)
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', type = str,  default = 'out')
    vcf_parser.add_argument('--out-dir', help = "Defines the output directory. Only used if multiple output files are created", default = 'Statistic_Files')

    # General arguments
    vcf_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')

    # Galaxy Option to pipe log to stdout
    vcf_parser.add_argument('--log-stdout', help = argparse.SUPPRESS, action = 'store_true')

    # Statistic based arguments
    statistic_list = ['weir-fst', 'windowed-weir-fst', 'TajimaD', 'site-pi', 'window-pi', 'freq', 'het-fit', 'het-fis', 'hardy-weinberg']

    vcf_parser.add_argument('--calc-statistic', metavar = metavar_list(statistic_list), help = 'The statistic to calculate', required = True, type = str, choices = statistic_list)

    # Statistic window options
    vcf_parser.add_argument('--statistic-window-size', help = 'Statistic window size', type = int)
    vcf_parser.add_argument('--statistic-window-step', help = 'Statistic window step size', type = int)

    # Position-based position filters
    vcf_parser.add_argument('--filter-include-positions', help = 'Defines a file of positions to include within a tsv file (chromosome and position)', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-positions', help = 'Defines a file of positions to exclude within a tsv file (chromosome and position)', action = parser_confirm_file())

    # BED-based position filters
    vcf_parser.add_argument('--filter-include-bed', help = 'Defines a BED file of positions to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-bed', help = 'Defines a BED file of positions to exclude', action = parser_confirm_file())

    # Non-model based filters
    vcf_parser.add_argument('--filter-include-indv', help = 'Defines the individual(s) to include. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-indv', help = 'Defines the individual(s) to exclude. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-include-indv-file', help = 'Defines a file of individuals to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-indv-file', help = 'Defines a file of individuals to exclude', action = parser_confirm_file())

    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()


def run (passed_arguments = []):
    '''
    Statistic calculation using VCFTools.

    This function uses the argparse-based function :py:func:`vcf_calc_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their VCFtools equivalent. Once all the
    VCFtools-based parameters are assigned, VCFtools is called.

    Parameters
    ----------
    --vcf : str
        Input VCF filename
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output filename prefix
    --out-dir : str
        Output directory
    --model-file : str
        Model filename
    --model : str
        Model to use
    --pop-file : str
        Defines a population file for relevant statistics. May be
        used multple times
    --calc-statistic : str
        Specifies the statistic to calculate
    --statistic-window-size : int
        Specifies the window size for window-based statistics
    --statistic-window-step : int
        Specifies step size between windows for specific window-based statistics
    --filter-include-indv : list or str
        Individual(s) to include. May be used multiple times. Not usable w/ --model
    --filter-exclude-indv : list or str
        Individual(s) to exclude. May be used multiple times. Not usable w/ --model
    --filter-include-indv-file : str
        File of individuals to include. Not usable w/ --model
    --filter-exclude-indv-file : str
        File of individuals to exclude.  Not usable w/ --model
    --filter-include-positions : str
        File of positions to include (tsv: chromosome and position)
    --filter-exclude-positions : str
        File of positions to exclude (tsv: chromosome and position)
    --filter-include-bed : str
        BED file of positions to include 
    --filter-exclude-bed : str
        BED file of positions to exclude
    
    Raises
    ------
    IOError
        Output file already exists and --overwrite is not specified
    Exception
        Incompatible arguments
    '''

    def calc_exception (selected_model, exc_type, exc_value, exc_traceback):

        # Check if the selected model was correctly assigned
        if selected_model: 
            logging.info('Delete temporary files due to exception')
            selected_model.delete_pop_files()
            selected_model.delete_ind_file()

        # Report the original error    
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    # Grab VCF arguments from command line
    vcf_args = vcf_calc_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_calc')

    # Argument container for vcftools
    vcftools_call_args = []

    # Used to store population information from either model or pop file(s)
    vcftools_pop_files = []

    # Checks if the user specified both a model and population files
    if vcf_args.model_file and vcf_args.pop_file:
        raise Exception('--model and --pop-file arguments are incompatible')

    # Check that individual-based filters are not being user with a model
    if vcf_args.model_file and (vcf_args.filter_include_indv or vcf_args.filter_exclude_indv):
        if vcf_args.filter_include_indv:
            raise Exception('--model and --filter-include-indv arguments are incompatible')
        if vcf_args.filter_exclude_indv:
            raise Exception('--model and --filter-exclude-indv arguments are incompatible')

    # Check that individuals-based filters are not being user with a model
    if vcf_args.model_file and (vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file):
        if vcf_args.filter_include_indv_file:
            raise Exception('--model and --filter_include_indv-file arguments are incompatible')
        if vcf_args.filter_exclude_indv_file:
            raise Exception('--model and --filter-exclude-indv-file arguments are incompatible')

    # Performs actions related to --model-file argument
    if vcf_args.model_file:

        # Check if a model has been specified
        if not vcf_args.model:
            raise IOError('No selected model. Please use --model to select a model')

        # Read in the model file
        models_in_file = read_model_file(vcf_args.model_file)

        # Check that the selected model was found in the file
        if vcf_args.model not in models_in_file:
            raise IOError('Selected model "%s" not found in: %s' % (vcf_args.model, vcf_args.model_file))

        # Select model, might change this in future versions
        selected_model = models_in_file[vcf_args.model]

        logging.info('%s assigned as model' % selected_model)

        # Check if the specified statistic is fst-based or fis-based
        if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst', 'het-fis']:

            # Create the population files
            selected_model.create_pop_files(file_ext = '.txt', overwrite = vcf_args.overwrite)

            # Store the population files
            vcftools_pop_files = selected_model.pop_files

        # Filter individuals for other statistics
        else:

            # Create individuals file
            selected_model.create_ind_file(overwrite = vcf_args.overwrite)

            # Assign the individuals file to vcftools
            vcftools_call_args.extend(['--keep', selected_model.ind_file])

        # Create a lambda function to pass the selected model along with the exception
        lambda_excepthook = lambda exc_type, exc_value, exc_traceback: calc_exception(selected_model, exc_type, exc_value, exc_traceback)

        # Set the special escape hook
        sys.excepthook = lambda_excepthook

    # Performs actions related to --pop-file argument
    if vcf_args.pop_file:

        # Check if the specified statistic is fst-based or fis-based
        if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst', 'het-fis']:

            # Store the population files
            vcftools_pop_files = vcf_args.pop_file

        # Return an error if --pop-file is not supported
        else:
            raise Exception('%s does not support the --pop-file argument' % vcf_args.calc_statistic)

    # Individuals-based filters
    if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
        # Used to include a file of individuals to keep
        if vcf_args.filter_exclude_indv_file:
            vcftools_call_args.extend(['--keep', vcf_args.filter_include_indv_file])

        # Used to include a file of individuals to remove
        if vcf_args.filter_exclude_indv_file:
            vcftools_call_args.extend(['--remove', vcf_args.filter_exclude_indv_file])

    # Individual-based filters
    if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
        if vcf_args.filter_include_indv:
            for indv_to_include in vcf_args.filter_include_indv:
                vcftools_call_args.extend(['--indv', indv_to_include])
        if vcf_args.filter_exclude_indv:
            for indv_to_exclude in vcf_args.filter_exclude_indv:
                vcftools_call_args.extend(['--remove-indv', indv_to_exclude])

    # Position (vcftools output file) filters
    if vcf_args.filter_include_positions or vcf_args.filter_exclude_positions:
        if vcf_args.filter_include_positions:
            vcftools_call_args.extend(['--positions', vcf_args.filter_include_positions])
        if vcf_args.filter_exclude_positions:
            vcftools_call_args.extend(['--exclude-positions', vcf_args.filter_exclude_positions])

    # Position (BED format file) filters
    if vcf_args.filter_include_bed or vcf_args.filter_exclude_bed:
        if vcf_args.filter_include_bed:
            vcftools_call_args.extend(['--bed', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            vcftools_call_args.extend(['--exclude-bed', vcf_args.filter_exclude_bed])

    # Check if windowed Fst is specified
    if vcf_args.calc_statistic == 'windowed-weir-fst':

        # Confirm that population files has been assigned
        if len(vcftools_pop_files) < 2:
            raise Exception('%s requires at least two populations to operate' % vcf_args.calc_statistic)

        # Confirm that the window size has been assigned
        if not vcf_args.statistic_window_size:
            raise Exception('%s requires a specifed window size to operate' % vcf_args.calc_statistic)

        # Confirm that window step size has been assigned
        if not vcf_args.statistic_window_step:
            raise Exception('%s requires a specifed window step size to operate' % vcf_args.calc_statistic)

        # Assigns the required window arguments
        vcftools_call_args.extend(['--fst-window-size', vcf_args.statistic_window_size, '--fst-window-step', vcf_args.statistic_window_step])

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.windowed.weir.fst'

    elif vcf_args.calc_statistic == 'weir-fst':

        # Confirm that population files has been assigned
        if len(vcftools_pop_files) < 2:
            raise Exception('%s requires at least two populations to operate' % vcf_args.calc_statistic)

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.weir.fst'

    elif vcf_args.calc_statistic == 'TajimaD':

        # Confirm that the window size has been assigned
        if not vcf_args.statistic_window_size:
            raise Exception('%s requires a specifed window size to operate' % vcf_args.calc_statistic)

        # Assigns all the vcftools arguments for calculating TajimaD
        vcftools_call_args.extend(['--TajimaD', vcf_args.statistic_window_size])

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.Tajima.D'

    elif vcf_args.calc_statistic == 'site-pi':

        # Assigns all the vcftools arguments for calculating pi
        vcftools_call_args.append('--site-pi')

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.sites.pi'

    elif vcf_args.calc_statistic == 'window-pi':

        # Confirm that the window size has been assigned
        if not vcf_args.statistic_window_size:
            raise Exception('%s requires a specifed window size to operate' % vcf_args.calc_statistic)

        # Confirm that window step size has been assigned
        if not vcf_args.statistic_window_step:
            raise Exception('%s requires a specifed window step size to operate' % vcf_args.calc_statistic)

        # Assigns all the vcftools arguments for calculating pi
        vcftools_call_args.extend(['--window-pi', vcf_args.statistic_window_size, '--window-pi-step', vcf_args.statistic_window_step])

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.windowed.pi'

    elif vcf_args.calc_statistic == 'hardy-weinberg':

        # Assigns all the vcftools arguments for the allele frequency
        vcftools_call_args.append('--hardy')

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.hwe'

    elif vcf_args.calc_statistic == 'freq':

        # Assigns all the vcftools arguments for the allele frequency
        vcftools_call_args.append('--freq')

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.frq'

    elif vcf_args.calc_statistic == 'het-fit':

        # Assigns all the vcftools arguments for calculating heterozygosity
        vcftools_call_args.append('--het')

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.het'

    elif vcf_args.calc_statistic == 'het-fis':

        # Confirm that at least one population file has been assigned
        if not vcftools_pop_files:
            raise Exception('%s requires at least one population to operate' % vcf_args.calc_statistic)

        # Assigns all the vcftools arguments for calculating heterozygosity
        vcftools_call_args.append('--het')

        # Assigns the suffix for the vcftools log file
        vcftools_out_suffix = '.het'

    logging.info('vcftools parameters assigned')

    # Check if the user has specified a output filename
    if vcf_args.out:

        # Assign the vcftools output filename, using the output filename
        vcftools_output_filename = vcf_args.out

    else:

        # Assign the vcftools output filename, using the prefix and suffix
        vcftools_output_filename = vcf_args.out_prefix + vcftools_out_suffix

    # Check if previous output should be overwritten
    if vcf_args.overwrite:

        # Check if the ouput will require the output directory
        if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst'] and len(vcftools_pop_files) > 2:
            
            # Check if the output directory is present
            if os.path.exists(vcf_args.out_dir):

                # Remove the output directory
                shutil.rmtree(vcf_args.out_dir)

    # If not to be overwritten, check if previous output exists
    else:

        # Confirm the vcftools output and log file do not exist
        check_for_vcftools_output(vcftools_output_filename)

        # Check if the ouput will require the output directory
        if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst'] and len(vcftools_pop_files) > 2:

            # Check if the output directory is present and report the error
            if os.path.exists(vcf_args.out_dir):
                raise IOError('Statistic Directory already exists')

    # Assigns the file argument for vcftools
    vcfname_arg = assign_vcftools_input_arg(vcf_args.vcf)

    logging.info('Input file assigned')

    # Run vcftools once if the statistic isn't het-fis
    if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst'] and len(vcftools_pop_files) > 2:

        def return_filename (filepath):
            return os.path.basename(filepath).split(os.extsep)[0]

        # Create the output directory
        if not os.path.exists(vcf_args.out_dir):
            os.makedirs(vcf_args.out_dir)

        # Loop each population file
        for first_pop_filepath, second_pop_filepath in itertools.combinations(vcftools_pop_files, 2):

            # Create the population-specific call
            pop_call_args = copy.deepcopy(vcftools_call_args)

            # Assign the population files
            pop_call_args.extend(['--weir-fst-pop', first_pop_filepath, '--weir-fst-pop', second_pop_filepath])

            # Extract filename from first filepath
            first_pop_filename = return_filename(first_pop_filepath)

            # Extract filename from second filepath
            second_pop_filename = return_filename(second_pop_filepath)

            # Create the population prefix, and join to the output directory
            pop_prefix = os.path.join(vcf_args.out_dir, vcf_args.out_prefix)

            # Update the population prefix with the population names
            pop_prefix += '.%s.%s' % (first_pop_filename, second_pop_filename)

            # The filename population file
            pop_filename = pop_prefix + vcftools_out_suffix

            # vcftools subprocess call, with stdout
            vcftools_err = call_vcftools(vcfname_arg + pop_call_args, output_filename = pop_filename)

            # Check if the log should be piped to the stdout
            if vcf_args.log_stdout:

                # Write the log to stdout
                sys.stdout.write(vcftools_err)

            # Check if log should be saved as a file
            else:

                # Produce the vcftools log file, in append mode
                produce_vcftools_log(vcftools_err, vcftools_output_filename, append_mode = True)

    elif vcf_args.calc_statistic == 'het-fis':

        # Loop each population file
        for vcftools_pop_file in vcftools_pop_files:

            # Create the population-specific call
            pop_call_args = vcftools_call_args + ['--keep', str(vcftools_pop_file)]

            # vcftools subprocess call
            vcftools_err = call_vcftools(vcfname_arg + pop_call_args, output_filename = vcftools_output_filename, append_mode = True)

            # Check if the log should be piped to the stdout
            if vcf_args.log_stdout:

                # Write the log to stdout
                sys.stdout.write(vcftools_err)

            # Check if log should be saved as a file
            else:

                # Produce the vcftools log file, in append mode
                produce_vcftools_log(vcftools_err, vcftools_output_filename, append_mode = True)

    else:

        # Check if either Fst statistic is specified
        if vcf_args.calc_statistic in ['windowed-weir-fst', 'weir-fst'] and len(vcftools_pop_files) == 2:

            # Assigns the population files to the vcftools call
            vcftools_call_args.extend([pop_args for pop_file in vcftools_pop_files for pop_args in ['--weir-fst-pop', pop_file]])

        # vcftools subprocess call
        vcftools_err = call_vcftools(vcfname_arg + vcftools_call_args, output_filename = vcftools_output_filename)

        # Check if the log should be piped to the stdout
        if vcf_args.log_stdout:

            # Write the log to stdout
            sys.stdout.write(vcftools_err)

        # Check if log should be saved as a file
        else:

            # Produce the vcftools log file
            produce_vcftools_log(vcftools_err, vcftools_output_filename)

    # Delete any files that were created for vcftools
    if vcf_args.model_file:
        selected_model.delete_pop_files()
        selected_model.delete_ind_file()

if __name__ == "__main__":
    initLogger()
    run()



import os
import sys
import argparse
import itertools
import copy
import shutil
import logging
import pandas as pd

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.logging_module import initLogger, logArgs

from pgpipe.picard import call_picard

def fasta_utility_parser(passed_arguments):
    '''fasta Argument Parser - Assigns arguments from command line'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction


    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    fasta_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    fasta_parser.add_argument('--fasta', help = "Input FASTA filename", type = str, nargs = '+', required = True, action = parser_confirm_file())

    # Utility based arguments
    utility_list = ['ref-dictionary']
    fasta_parser.add_argument('--utility', metavar = metavar_list(utility_list), help = 'The utility to use', type = str, choices = utility_list)

    # Output file arguments
    out_format_list = ['fasta', 'fasta.gz']
    out_format_default = 'fasta.gz'

    fasta_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Specifies the format of FASTA-based output', type = str, choices = out_format_list, default = out_format_default)


    fasta_parser.add_argument('--out', help = 'Output filename. If used, overrides --out-prefix', type = str)
    fasta_parser.add_argument('--out-prefix', help = 'Defines the output prefix', default = 'out')

    fasta_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')

    fasta_parser.add_argument('--picard-path', help = "Defines path to locate picard.jar", type = str)

    if passed_arguments:
        return fasta_parser.parse_args(passed_arguments)
    else:
        return fasta_parser.parse_args()

def run (passed_arguments = []):
    '''
    Liftover for VCF-formatted files

    Automates the liftover function from picard for VCF-formatted files.

    Parameters
    ----------
    --fasta : str
        Input FASTA filename
	--utility : str
        Specifies the utility to be used. Choices: ref-dictionary
    --out-format : str
        Format of the output file
    --out-prefix : str
        Output filename prefix
    --out : str
        Output filename. Overrides --out-prefix
    --overwrite
        If previous output should be overwriten

    Returns
    -------
    output : file
        Statistic file output
    log : file
        Log file output

    Raises
    ------
    IOError
        Input FASTA file does not exist
    IOError
        Output file already exists
    '''

    # Grab FASTA arguments from command line
    fasta_args = fasta_utility_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(fasta_args, func_name = 'fasta_utilities')

    # Check if a reference dictionary was not assigned
    if fasta_args.utility == 'ref-dictionary':

        # Create arg list for sequence dictionary
        picard_seqdict_call_args = ['CreateSequenceDictionary']

        # Add the reference file to the sequence dictionary args
        picard_seqdict_call_args.append('R=' + fasta_args.fasta)

        # Create dictionary for fasta reference
        call_picard(fasta_args.picard_path, picard_seqdict_call_args)


if __name__ == "__main__":
    #initLogger()
    run()


import os
import sys
import random
import argparse
import pysam
import itertools
import logging
import shutil
import copy
import numpy as np
import pandas as pd
from collections import defaultdict

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

# Import log initializer
from pgpipe.logging_module import initLogger, logArgs

# Import basic vcftools functions
from pgpipe.vcftools import *

# Import basic vcf
from pgpipe.bcftools import check_for_index, create_index

# Model file related functions
from pgpipe.model import read_model_file

def sampler_parser(passed_arguments):
    '''Sampler Argument Parser - Assigns arguments from command line.'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string = None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    sampler_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments
    sampler_parser.add_argument('--vcf', help = "Input VCF filename", type = str, required = True, action = parser_confirm_file())

    # Model file arguments.
    sampler_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    sampler_parser.add_argument('--model', help = 'Defines the model to analyze', type = str)

    # Other input arguments.
    sampler_parser.add_argument('--statistic-file', help='Specifies the statistic file for filtering', required = True, type = str, action = parser_confirm_file())
    sampler_parser.add_argument("--vcf-index", help = "VCF Index filename", type = str, action = parser_confirm_file())

    # Output arguents
    sampler_parser.add_argument('--sample-file', help = 'Specifies the sampled (statistic file) tsv output filename', type = str, default = 'sampled_data.tsv')

    sampler_parser.add_argument('--out-dir', help = 'Specifies the VCF output directory', type = str, default = 'Sample_Files')
    sampler_parser.add_argument('--out-prefix', help = 'Specifies the VCF output filename prefix', type = str, default = 'Sample')

    out_format_list = ['vcf', 'vcf.gz', 'bcf']
    out_format_default = 'vcf.gz'

    sampler_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Specifies the output format.', type = str, choices = out_format_list, default = out_format_default)

    # General arguments
    sampler_parser.add_argument('--overwrite', help = "Specifies if previous output files should be overwritten", action = 'store_true')

    # Statistic based arguments.
    statistic_list = ['windowed-weir-fst', 'TajimaD', 'window-pi']
    statistic_default = 'windowed-weir-fst'
    sampler_parser.add_argument('--calc-statistic', metavar = metavar_list(statistic_list), help = 'Specifies the statistic calculated ', type=str, choices = statistic_list, default = statistic_default)

    sampler_parser.add_argument('--statistic-window-size', help = 'Specifies the size of window calculations (if BIN_END is absent)', type = int, default = 10000)

    # Sampling methods. Currently mutually exclusive to only allow a single sampling method
    sampling_list = ['uniform', 'random']
    sampling_default = 'random'
    sampler_parser.add_argument('--sampling-scheme', metavar = metavar_list(sampling_list), help = 'Specifies the sampling scheme ', type=str, choices = sampling_list, default = sampling_default)

    # Sampling options
    sampler_parser.add_argument('--uniform-bins', help="Number of bins in uniform sampling", type = int, default = 10)
    sampler_parser.add_argument('--sample-size', help="Total sample size. If using the uniform sampling scheme, this number must be divisible by the bins", type = int, default = 200)

    # Random options
    sampler_parser.add_argument('--random-seed', help="Defines the random seed value for the random number generator", type = int, default = random.randint(1, 1000000000))

    # Individual filters
    sampler_parser.add_argument('--filter-keep', help = 'Specifies a file of individuals to keep', action = parser_confirm_file())
    sampler_parser.add_argument('--filter-remove', help = 'Specifies a file of individuals to remove', action = parser_confirm_file())

    # Position-based position filters
    sampler_parser.add_argument('--filter-include-positions', help = 'Specifies a set of sites to include within a file (tsv chromosome and position)', action = parser_confirm_file())
    sampler_parser.add_argument('--filter-exclude-positions', help = 'Specifies a set of sites to exclude within a file (tsv chromosome and position)', action = parser_confirm_file())

    if passed_arguments:
        return sampler_parser.parse_args(passed_arguments)
    else:
        return sampler_parser.parse_args()

def random_vcftools_sampler (vcftools_samples, sample_size):
    '''Random Sampler. Returns a list of randomly selected elements from
    vcftools_samples. The length of this list is defined by sample_size.'''

    try:
        return list(np.random.choice(vcftools_samples, int(sample_size), replace = False))
    except ValueError:
        return vcftools_samples

def uniform_vcftools_sampler (vcftools_samples, uniform_bins, sample_size):
    '''Uniform Sampler. Seperates the elements of vcftools_samples into
    equal sized bins (number of bins specified by .uniform_bins arg). The
    samples in the bins are then randomly selected (number of selected
    samples is defined by: .sample_size / .uniform_bins).'''

    binned_sample_database = defaultdict(list)

    binned_samples = pd.cut(vcftools_samples, uniform_bins, labels = False)

    for vcftools_sample, sample_bin in zip(range(len(vcftools_samples)), binned_samples):
        binned_sample_database[sample_bin].append(vcftools_sample)

    uniform_samples = []

    for current_samples in binned_sample_database.values():
        uniform_samples.extend(random_vcftools_sampler(current_samples, sample_size / uniform_bins))

    return uniform_samples

def assign_position_columns (sample_headers):
    '''Column assignment. Returns the columns that contain the following
    headers: CHROM, BIN_START, BIN_END. Will also produce an error if not
    all the columns were found.'''

    if not all(pos_header in sample_headers for pos_header in ['CHROM', 'BIN_START', 'BIN_END']):
        # Assign missing headers
        missing_headers = [pos_header for pos_header in ['CHROM', 'BIN_START', 'BIN_END'] if pos_header not in sample_headers]
        # Print error message with missing headers
        raise ValueError('Cannot find %s column(s) in file specified by --statistic-file.' % ', '.join(missing_headers))

    return sample_headers.index('CHROM'), sample_headers.index('BIN_START'), sample_headers.index('BIN_END')

def assign_statistic_column (sample_headers, statistic):
    statistic_converter = {'windowed-weir-fst':'MEAN_FST', 'TajimaD':'TajimaD', 'window-pi':'PI'}

    if statistic not in statistic_converter:
        raise Exception('Statistic not found. Statistic list needs to be updated. Please contact the PPP Team.')

    if statistic_converter[statistic] not in sample_headers:
        raise ValueError('Statistic selected not found in file specified by --statistic-file.')

    return statistic_converter[statistic]

def run (passed_arguments = []):
    '''
        Statistics sampler for VCFTools output.

        Automates the sampling process of a specified statistic from VCFTools
        output files. The function allows the user to select both the statistic
        in question (windowed Fst [default] and Tajima's D) and the sampling
        scheme(random [default] and uniform).

        Parameters
        ----------
        --vcf : str
            Specifies the input VCF filename
        --out : str
            Specifies the VCF output filename
        --sample-file : str
            Specifies the sampled (statistic file) tsv output filename
        --statistic-file : str
            Specifies the statistic file for filtering
        --calc-statistic : str
            Specifies the statistic to calculate. Choices: windowed-weir-fst
            (default) and TajimaD
        --statistic-window-size : int
            Specifies the window size of the statistic. Required for Tajima's D
        --sampling-scheme : str
            Specifies the sampling scheme to use. Choices: random (default) and
            uniform
        --uniform-bins : int
            Specifies the number of bins for the uniform sampler
        --sample-size : int
            Specifies the total sample size. Note: If using the uniform sampling
            scheme, this number must be divisible by number of uniform bins
        --random-seed : int
            Specifies the random seed value for the random number generator

        Returns
        -------
        output : file
            Sampled statistic file
        samples : file
            Sampled VCF output

        Raises
        ------
        IOError
            Input VCF file does not exist
        IOError
            Output file already exists
        ValueError
            Sample size larger than the number of samples
        ValueError
            Sample size not divisible by the bin count
        ValueError
            Window size argument conflicts with values
        TypeError
            Window size argument not defined (if necessary)
    '''

    # Get arguments from command line
    sampler_args = sampler_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(sampler_args, func_name = 'vcf_sampler')

    # Set the random seed
    np.random.seed(sampler_args.random_seed)

    # Check if user has selected to overwrite previous output
    if not sampler_args.overwrite:
        # Check if output dir already exists
        if os.path.exists(sampler_args.out_dir):
            raise IOError('Sample Directory already exists')

        # Check if output file already exists
        if os.path.isfile(sampler_args.sample_file):
            raise IOError('Sample file already exists')
    else:
        # If previous output is to be overwritten, the previous output directory needs to be removed
        if os.path.exists(sampler_args.out_dir):
            shutil.rmtree(sampler_args.out_dir)

    # Read in the sample file
    vcftools_samples = pd.read_csv(sampler_args.statistic_file, sep = '\t')

    # Confirm statistic is correct, and assign the column ID
    assigned_statistic = assign_statistic_column(list(vcftools_samples), sampler_args.calc_statistic)

    # Remove NaN datapoints for the statistic
    vcftools_samples = vcftools_samples.dropna(subset=[assigned_statistic])

    logging.info('Sample (i.e. statistic) file assigned')

    # Confirm there are enough samples
    if len(vcftools_samples) < sampler_args.sample_size:
        raise ValueError('Sample size larger than the number of samples within sample file')

    # UPDATE Planned
    # Improve from equal to too few samples (i.e. samples = 100, sample_size = 95)
    elif len(vcftools_samples) == sampler_args.sample_size:
        # Warns the user of poor sampling
        logging.warning('Sample size equal to number of samples within sample file')

    # Run the uniform sampler
    if sampler_args.sampling_scheme == 'uniform':
        if sampler_args.sample_size % sampler_args.uniform_bins != 0:
            raise ValueError('Sample size not divisible by the bin count')

        selected_samples = sorted(uniform_vcftools_sampler(list(vcftools_samples[assigned_statistic]), sampler_args.uniform_bins, sampler_args.sample_size))

        if len(selected_samples) < sampler_args.sample_size:
            logging.warning('Uniform sampler was unable to reach sample size. Likely too few samples per bin, consider reducing the number of bins')

        logging.info('Uniform sampling complete')

    # Run the random sampler
    if sampler_args.sampling_scheme == 'random':

        selected_samples = sorted(random_vcftools_sampler(list(vcftools_samples.index), sampler_args.sample_size))

        logging.info('Random sampling complete')

    # Reduce to only selected samples
    sampled_samples = vcftools_samples[vcftools_samples.index.isin(selected_samples)].copy()

    # Create selected samples TSV file, with either the default filename or a user-defined filename
    sampled_samples.to_csv(sampler_args.sample_file, sep = '\t', float_format = '%g')

    logging.info('Created selected samples file')

    # Create the vcf/bcf output directory
    if not os.path.exists(sampler_args.out_dir):
        os.makedirs(sampler_args.out_dir)

    # Check if the data was sampled using Tajima's D. Adds BIN_END columns
    # that is required to fetch intervals from vcf files. Will need to be
    # edited when we add more statistics.
    if sampler_args.calc_statistic == 'TajimaD':
        # Check that the window size used in vcf_calc has been defined
        if not sampler_args.statistic_window_size:
            raise TypeError("--statistic-window-size argument required for the Tajima's D statistic")

        # Create iterator of all unique bin combinations
        bin_start_iter = itertools.combinations(list(sampled_samples['BIN_START']), 2)

        # Check if the bin combinations are divisible by the window size (to check if the window size is likely correct)
        if not all([abs(bin_1 - bin_2) % sampler_args.statistic_window_size == 0  for bin_1, bin_2 in bin_start_iter]):
            raise ValueError("--statistic-window-size argument conflicts with values in sample file")

        # Create list of bin start positions
        bin_start_list = list(sampled_samples['BIN_START'])

        # Create list of bin end positions
        bin_end_list = [start_pos + (sampler_args.statistic_window_size - 1) for start_pos in bin_start_list]

        # Add bin end positions to sampled_samples
        sampled_samples['BIN_END'] = pd.Series(bin_end_list, index = sampled_samples.index)

    # Open the VCF file
    if sampler_args.vcf_index:
        # Read in the file using pysam
        vcf_input = pysam.VariantFile(sampler_args.vcf, index_filename = sampler_args.vcf_index)
    else:
        # Check if there is an index file
        if check_for_index(sampler_args.vcf) == False:
            # Create an index if not found
            create_index(sampler_args.vcf)
        # Read in the file using pysam
        vcf_input = pysam.VariantFile(sampler_args.vcf)

    # Check if the user has specified a model file
    if sampler_args.model_file and sampler_args.model:

        # Read in the models
        models_in_file = read_model_file(sampler_args.model_file)

        # Check that the selected model was not found in the file
        if sampler_args.model not in models_in_file:
            raise IOError('Selected model "%s" not found in: %s' % (sampler_args.model, sampler_args.model_file))

        # Select model, might change this in future versions
        selected_model = models_in_file[sampler_args.model]

        # List of the individuals within the vcf
        individuals_in_file = list(vcf_input.header.samples)

        if not set(selected_model.inds).issubset(set(individuals_in_file)):
            raise IOError('Individuals differ between model "%s" and input.' % sampler_args.model)

        # Reduce individuals to those in model
        vcf_input.subset_samples(selected_model.inds)

    # If no model was specifed, check if a keep or remove file was specifed
    elif sampler_args.filter_keep or sampler_args.filter_remove:

        # List of the individuals within the vcf
        individuals_in_file = list(vcf_input.header.samples)

        # Check if a keep file was specifed
        if sampler_args.filter_keep:

            # List of the individuals to keep
            selected_individuals = []

            # Read in the individuals within the keep file
            with open(sampler_args.filter_keep, 'rU') as keep_file:
                for keep_line in keep_file:

                    if keep_line.strip() not in individuals_in_file:
                        raise IOError('Individuals differ between --keep-file and input.')

                    selected_individuals.append(keep_line.strip())

        # Check if a remove file was specifed
        elif sampler_args.filter_remove:

            # List of the individuals to keep
            selected_individuals = copy.deepcopy(individuals_in_file)

            # Read in the individuals within the keep file
            with open(sampler_args.filter_remove, 'rU') as remove_file:
                for remove_line in remove_file:

                    if remove_line.strip() not in individuals_in_file:
                        raise IOError('Individuals differ between --remove-file and input.')

                    selected_individuals.remove(remove_line.strip())

        # Reduce individuals to those in model
        vcf_input.subset_samples(selected_individuals)

    # Sites to be included
    include_sites =  pd.DataFrame()

    # Check if the user has specified an include sites file
    if sampler_args.filter_include_positions:
        include_sites = pd.read_csv(sampler_args.filter_include_positions, sep = '\t')

    # Sites to be excluded
    exclude_sites = pd.DataFrame()

    # Check if the user has specified an exclude sites file
    if sampler_args.filter_exclude_positions:
        exclude_sites = pd.read_csv(sampler_args.filter_exclude_positions, sep = '\t')

    # Get the chromosome, start, and end columns
    chr_col, start_col, end_col = assign_position_columns(list(sampled_samples))

    # iterate the selected samples
    for sampled_count, sampled_row in enumerate(sampled_samples.values):

        # Assign filename for sample.
        if sampler_args.out_format == 'vcf.gz':
            sample_filename =  sampler_args.out_prefix + '_%s.' %sampled_count + 'vcf'
        else:
            sample_filename =  sampler_args.out_prefix + '_%s.' %sampled_count + sampler_args.out_format

        # Join the output directory and the sample filename paths
        sample_path = os.path.join(sampler_args.out_dir, sample_filename)

        # Create the VCF output file, with either the default filename or a user-defined filename
        vcf_output = pysam.VariantFile(sample_path, 'w', header = vcf_input.header)

        # Fetch positions specified from the vcf input file
        for vcf_record in vcf_input.fetch(sampled_row[chr_col], int(sampled_row[start_col]), int(sampled_row[end_col])):

            # Bool to determine if the record should be saved
            save_record = True

            # Check if the user has specified an include sites file
            if sampler_args.filter_include_positions:
                # Check if the current record is not an included site
                if not ((include_sites['CHROM'] == vcf_record.chrom) & (include_sites['POS'] == vcf_record.pos)).any():
                    save_record = False

            # Check if the user has specified an exclude sites file
            if sampler_args.filter_exclude_positions:
                # Check if the current record is an excluded site
                if ((exclude_sites['CHROM'] == vcf_record.chrom) & (exclude_sites['POS'] == vcf_record.pos)).any():
                    save_record = False

            # Check if the record is to be saved
            if save_record:
                vcf_output.write(vcf_record)

        vcf_output.close()

        # Compress vcf to vcf.gz using bgzip
        if sampler_args.out_format == 'vcf.gz':
            bgzip_compress_vcf(sample_path)

    vcf_input.close()


if __name__ == "__main__":
    initLogger()
    run()
